from __future__ import annotations

from collections.abc import Mapping
from typing import Any

import jax.numpy as jnp
import numpy as np
import scipy.stats as ss
from tjax import JaxRealArray, NumpyComplexArray, NumpyRealArray, abs_square, create_diagonal_array
from typing_extensions import override

from efax import (BernoulliEP, BernoulliNP, BetaEP, BetaNP, ChiEP, ChiNP, ChiSquareEP, ChiSquareNP,
                  ComplexCircularlySymmetricNormalEP, ComplexCircularlySymmetricNormalNP,
                  ComplexMultivariateUnitNormalEP, ComplexMultivariateUnitNormalNP, ComplexNormalEP,
                  ComplexNormalNP, ComplexUnitNormalEP, ComplexUnitNormalNP, DirichletEP,
                  DirichletNP, ExponentialEP, ExponentialNP, GammaEP, GammaNP,
                  GeneralizedDirichletEP, GeneralizedDirichletNP, GeometricEP, GeometricNP,
                  IsotropicNormalEP, IsotropicNormalNP, JointDistributionE, JointDistributionN,
                  LogarithmicEP, LogarithmicNP, MultivariateDiagonalNormalEP,
                  MultivariateDiagonalNormalNP, MultivariateFixedVarianceNormalEP,
                  MultivariateFixedVarianceNormalNP, MultivariateNormalEP, MultivariateNormalNP,
                  MultivariateUnitNormalEP, MultivariateUnitNormalNP, NegativeBinomialEP,
                  NegativeBinomialNP, NormalEP, NormalNP, PoissonEP, PoissonNP, RayleighEP,
                  RayleighNP, ScipyComplexMultivariateNormal, ScipyComplexNormal, ScipyDirichlet,
                  ScipyGeneralizedDirichlet, ScipyGeometric, ScipyJointDistribution,
                  ScipyMultivariateNormal, ScipyVonMises, ScipyVonMisesFisher, Structure,
                  SubDistributionInfo, UnitNormalEP, UnitNormalNP, VonMisesFisherEP,
                  VonMisesFisherNP, WeibullEP, WeibullNP)

from .distribution_info import DistributionInfo


class BernoulliInfo(DistributionInfo[BernoulliNP, BernoulliEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: BernoulliEP) -> Any:
        return ss.bernoulli(p.probability)

    @override
    def exp_class(self) -> type[BernoulliEP]:
        return BernoulliEP

    @override
    def nat_class(self) -> type[BernoulliNP]:
        return BernoulliNP


class GeometricInfo(DistributionInfo[GeometricNP, GeometricEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: GeometricEP) -> Any:
        # Scipy uses a different definition geometric distribution.  The parameter p is inverse
        # odds.
        return ScipyGeometric(np.asarray(1.0 / (1.0 + p.mean)))

    @override
    def scipy_to_exp_family_observation(self, x: NumpyRealArray) -> JaxRealArray:
        return jnp.asarray(x - 1)

    @override
    def exp_class(self) -> type[GeometricEP]:
        return GeometricEP

    @override
    def nat_class(self) -> type[GeometricNP]:
        return GeometricNP


class PoissonInfo(DistributionInfo[PoissonNP, PoissonEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: PoissonEP) -> Any:
        return ss.poisson(p.mean)

    @override
    def exp_class(self) -> type[PoissonEP]:
        return PoissonEP

    @override
    def nat_class(self) -> type[PoissonNP]:
        return PoissonNP


class NegativeBinomialInfo(DistributionInfo[NegativeBinomialNP, NegativeBinomialEP,
                                            NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: NegativeBinomialEP) -> Any:
        return ss.nbinom(p.failures, 1.0 / (1.0 + p.mean / p.failures))

    @override
    def exp_class(self) -> type[NegativeBinomialEP]:
        return NegativeBinomialEP

    @override
    def nat_class(self) -> type[NegativeBinomialNP]:
        return NegativeBinomialNP


class LogarithmicInfo(DistributionInfo[LogarithmicNP, LogarithmicEP, NumpyRealArray]):
    @override
    def nat_to_scipy_distribution(self, q: LogarithmicNP) -> Any:
        return ss.logser(np.exp(q.log_probability))

    @override
    def exp_class(self) -> type[LogarithmicEP]:
        return LogarithmicEP

    @override
    def nat_class(self) -> type[LogarithmicNP]:
        return LogarithmicNP


class NormalInfo(DistributionInfo[NormalNP, NormalEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: NormalEP) -> Any:
        return ss.norm(p.mean, np.sqrt(p.variance()))

    @override
    def exp_class(self) -> type[NormalEP]:
        return NormalEP

    @override
    def nat_class(self) -> type[NormalNP]:
        return NormalNP


class UnitNormalInfo(DistributionInfo[UnitNormalNP, UnitNormalEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: UnitNormalEP) -> Any:
        return ss.norm(p.mean, 1.0)

    @override
    def exp_class(self) -> type[UnitNormalEP]:
        return UnitNormalEP

    @override
    def nat_class(self) -> type[UnitNormalNP]:
        return UnitNormalNP


class MultivariateFixedVarianceNormalInfo(DistributionInfo[MultivariateFixedVarianceNormalNP,
                                                           MultivariateFixedVarianceNormalEP,
                                                           NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: MultivariateFixedVarianceNormalEP) -> Any:
        cov = np.tile(np.eye(p.dimensions()), (*p.shape, 1, 1))
        for i in np.ndindex(*p.shape):
            cov[i] *= p.variance[i]
        return ScipyMultivariateNormal.from_mc(mean=np.asarray(p.mean), cov=np.asarray(cov))

    @override
    def exp_class(self) -> type[MultivariateFixedVarianceNormalEP]:
        return MultivariateFixedVarianceNormalEP

    @override
    def nat_class(self) -> type[MultivariateFixedVarianceNormalNP]:
        return MultivariateFixedVarianceNormalNP


class MultivariateUnitNormalInfo(DistributionInfo[MultivariateUnitNormalNP,
                                                  MultivariateUnitNormalEP,
                                                  NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: MultivariateUnitNormalEP) -> Any:
        return ScipyMultivariateNormal.from_mc(mean=np.asarray(p.mean))

    @override
    def exp_class(self) -> type[MultivariateUnitNormalEP]:
        return MultivariateUnitNormalEP

    @override
    def nat_class(self) -> type[MultivariateUnitNormalNP]:
        return MultivariateUnitNormalNP


class IsotropicNormalInfo(DistributionInfo[IsotropicNormalNP, IsotropicNormalEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: IsotropicNormalEP) -> Any:
        v = p.variance()
        e = np.eye(self.dimensions)
        return ScipyMultivariateNormal.from_mc(mean=np.asarray(p.mean),
                                               cov=np.asarray(np.multiply.outer(v, e)))

    @override
    def exp_class(self) -> type[IsotropicNormalEP]:
        return IsotropicNormalEP

    @override
    def nat_class(self) -> type[IsotropicNormalNP]:
        return IsotropicNormalNP


class MultivariateDiagonalNormalInfo(DistributionInfo[MultivariateDiagonalNormalNP,
                                                      MultivariateDiagonalNormalEP,
                                                      NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: MultivariateDiagonalNormalEP) -> Any:
        variance = np.asarray(p.variance())
        covariance = create_diagonal_array(variance)
        return ScipyMultivariateNormal.from_mc(mean=np.asarray(p.mean), cov=covariance)

    @override
    def exp_class(self) -> type[MultivariateDiagonalNormalEP]:
        return MultivariateDiagonalNormalEP

    @override
    def nat_class(self) -> type[MultivariateDiagonalNormalNP]:
        return MultivariateDiagonalNormalNP


class MultivariateNormalInfo(DistributionInfo[MultivariateNormalNP, MultivariateNormalEP,
                                              NumpyRealArray]):
    def __init__(self, dimensions: int) -> None:
        super().__init__(dimensions=dimensions, safety=0.1)

    @override
    def exp_to_scipy_distribution(self, p: MultivariateNormalEP) -> Any:
        # Correct numerical errors introduced by various conversions.
        mean = np.asarray(p.mean, dtype=np.float64)
        v = np.asarray(p.variance(), dtype=np.float64)
        v_transpose = v.swapaxes(-1, -2)
        covariance = np.tril(v) + np.triu(v_transpose, 1)
        return ScipyMultivariateNormal.from_mc(mean=mean, cov=covariance)

    @override
    def exp_class(self) -> type[MultivariateNormalEP]:
        return MultivariateNormalEP

    @override
    def nat_class(self) -> type[MultivariateNormalNP]:
        return MultivariateNormalNP


class ComplexUnitNormalInfo(DistributionInfo[ComplexUnitNormalNP, ComplexUnitNormalEP,
                                             NumpyComplexArray]):
    @override
    def exp_to_scipy_distribution(self, p: ComplexUnitNormalEP) -> Any:
        mean = np.asarray(p.mean, dtype=np.complex128)
        variance = np.ones_like(np.real(mean))
        pseudo_variance = np.zeros_like(mean)
        return ScipyComplexNormal(mean, variance, pseudo_variance)

    @override
    def exp_class(self) -> type[ComplexUnitNormalEP]:
        return ComplexUnitNormalEP

    @override
    def nat_class(self) -> type[ComplexUnitNormalNP]:
        return ComplexUnitNormalNP


class ComplexNormalInfo(DistributionInfo[ComplexNormalNP, ComplexNormalEP, NumpyComplexArray]):
    @override
    def exp_to_scipy_distribution(self, p: ComplexNormalEP) -> Any:
        mean = np.asarray(p.mean, dtype=np.complex128)
        second_moment = np.asarray(p.second_moment, dtype=np.float64)
        pseudo_second_moment = np.asarray(p.pseudo_second_moment, dtype=np.complex128)
        return ScipyComplexNormal(mean,
                                  second_moment - abs_square(mean),
                                  pseudo_second_moment - np.square(mean))

    @override
    def exp_class(self) -> type[ComplexNormalEP]:
        return ComplexNormalEP

    @override
    def nat_class(self) -> type[ComplexNormalNP]:
        return ComplexNormalNP


class ComplexMultivariateUnitNormalInfo(DistributionInfo[ComplexMultivariateUnitNormalNP,
                                                         ComplexMultivariateUnitNormalEP,
                                                         NumpyComplexArray]):
    @override
    def exp_to_scipy_distribution(self, p: ComplexMultivariateUnitNormalEP) -> Any:
        return ScipyComplexMultivariateNormal(mean=np.asarray(p.mean))

    @override
    def exp_class(self) -> type[ComplexMultivariateUnitNormalEP]:
        return ComplexMultivariateUnitNormalEP

    @override
    def nat_class(self) -> type[ComplexMultivariateUnitNormalNP]:
        return ComplexMultivariateUnitNormalNP


class ComplexCircularlySymmetricNormalInfo(DistributionInfo[ComplexCircularlySymmetricNormalNP,
                                                            ComplexCircularlySymmetricNormalEP,
                                                            NumpyComplexArray]):
    @override
    def exp_to_scipy_distribution(self, p: ComplexCircularlySymmetricNormalEP) -> Any:
        return ScipyComplexMultivariateNormal(variance=np.asarray(p.variance))

    @override
    def exp_class(self) -> type[ComplexCircularlySymmetricNormalEP]:
        return ComplexCircularlySymmetricNormalEP

    @override
    def nat_class(self) -> type[ComplexCircularlySymmetricNormalNP]:
        return ComplexCircularlySymmetricNormalNP


class ExponentialInfo(DistributionInfo[ExponentialNP, ExponentialEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: ExponentialEP) -> Any:
        return ss.expon(0, p.mean)

    @override
    def exp_class(self) -> type[ExponentialEP]:
        return ExponentialEP

    @override
    def nat_class(self) -> type[ExponentialNP]:
        return ExponentialNP


class RayleighInfo(DistributionInfo[RayleighNP, RayleighEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: RayleighEP) -> Any:
        return ss.rayleigh(scale=np.sqrt(p.chi / 2.0))

    @override
    def exp_class(self) -> type[RayleighEP]:
        return RayleighEP

    @override
    def nat_class(self) -> type[RayleighNP]:
        return RayleighNP


class BetaInfo(DistributionInfo[BetaNP, BetaEP, NumpyRealArray]):
    def __init__(self) -> None:
        super().__init__(dimensions=2, safety=0.1)

    @override
    def nat_to_scipy_distribution(self, q: BetaNP) -> Any:
        n1 = q.alpha_minus_one + 1.0
        return ss.beta(n1[..., 0], n1[..., 1])

    @override
    def exp_class(self) -> type[BetaEP]:
        return BetaEP

    @override
    def nat_class(self) -> type[BetaNP]:
        return BetaNP


class GammaInfo(DistributionInfo[GammaNP, GammaEP, NumpyRealArray]):
    @override
    def nat_to_scipy_distribution(self, q: GammaNP) -> Any:
        shape = q.shape_minus_one + 1.0
        scale = -1.0 / q.negative_rate
        return ss.gamma(shape, scale=scale)

    @override
    def exp_class(self) -> type[GammaEP]:
        return GammaEP

    @override
    def nat_class(self) -> type[GammaNP]:
        return GammaNP


class DirichletInfo(DistributionInfo[DirichletNP, DirichletEP, NumpyRealArray]):
    def __init__(self, dimensions: int) -> None:
        super().__init__(dimensions=dimensions, safety=0.1)

    @override
    def nat_to_scipy_distribution(self, q: DirichletNP) -> Any:
        return ScipyDirichlet(np.asarray(q.alpha_minus_one, dtype=np.float64) + 1.0)

    @override
    def scipy_to_exp_family_observation(self, x: NumpyRealArray) -> JaxRealArray:
        return jnp.asarray(x[..., : -1])

    @override
    def exp_class(self) -> type[DirichletEP]:
        return DirichletEP

    @override
    def nat_class(self) -> type[DirichletNP]:
        return DirichletNP


class GeneralizedDirichletInfo(DistributionInfo[GeneralizedDirichletNP, GeneralizedDirichletEP,
                                                NumpyRealArray]):
    def __init__(self, dimensions: int) -> None:
        super().__init__(dimensions=dimensions, safety=3.0)

    @override
    def nat_to_scipy_distribution(self, q: GeneralizedDirichletNP) -> Any:
        alpha, beta = q.alpha_beta()
        return ScipyGeneralizedDirichlet(np.asarray(alpha), np.asarray(beta))

    @override
    def exp_class(self) -> type[GeneralizedDirichletEP]:
        return GeneralizedDirichletEP

    @override
    def nat_class(self) -> type[GeneralizedDirichletNP]:
        return GeneralizedDirichletNP


class VonMisesInfo(DistributionInfo[VonMisesFisherNP, VonMisesFisherEP, NumpyRealArray]):
    def __init__(self) -> None:
        super().__init__(dimensions=2)

    @override
    def nat_to_scipy_distribution(self, q: VonMisesFisherNP) -> Any:
        kappa, angle = q.to_kappa_angle()
        return ScipyVonMises(np.asarray(kappa), np.asarray(angle))

    @override
    def scipy_to_exp_family_observation(self, x: NumpyRealArray) -> JaxRealArray:
        result = np.empty((*x.shape, 2))
        result[..., 0] = np.cos(x)
        result[..., 1] = np.sin(x)
        return jnp.asarray(result)

    @override
    def exp_class(self) -> type[VonMisesFisherEP]:
        return VonMisesFisherEP

    @override
    def nat_class(self) -> type[VonMisesFisherNP]:
        return VonMisesFisherNP


class VonMisesFisherInfo(DistributionInfo[VonMisesFisherNP, VonMisesFisherEP, NumpyRealArray]):
    @override
    def nat_to_scipy_distribution(self, q: VonMisesFisherNP) -> Any:
        kappa = np.asarray(q.kappa())
        mu = np.asarray(q.mean_times_concentration) / kappa[..., np.newaxis]
        return ScipyVonMisesFisher(mu, kappa)

    @override
    def exp_class(self) -> type[VonMisesFisherEP]:
        return VonMisesFisherEP

    @override
    def nat_class(self) -> type[VonMisesFisherNP]:
        return VonMisesFisherNP


class ChiSquareInfo(DistributionInfo[ChiSquareNP, ChiSquareEP, NumpyRealArray]):
    @override
    def nat_to_scipy_distribution(self, q: ChiSquareNP) -> Any:
        return ss.chi2((q.k_over_two_minus_one + 1.0) * 2.0)

    @override
    def exp_class(self) -> type[ChiSquareEP]:
        return ChiSquareEP

    @override
    def nat_class(self) -> type[ChiSquareNP]:
        return ChiSquareNP


class ChiInfo(DistributionInfo[ChiNP, ChiEP, NumpyRealArray]):
    @override
    def nat_to_scipy_distribution(self, q: ChiNP) -> Any:
        return ss.chi((q.k_over_two_minus_one + 1.0) * 2.0)

    @override
    def exp_class(self) -> type[ChiEP]:
        return ChiEP

    @override
    def nat_class(self) -> type[ChiNP]:
        return ChiNP


class WeibullInfo(DistributionInfo[WeibullNP, WeibullEP, NumpyRealArray]):
    @override
    def exp_to_scipy_distribution(self, p: WeibullEP) -> Any:
        scale = p.chi ** (1.0 / p.concentration)
        return ss.weibull_min(p.concentration, scale=scale)

    @override
    def exp_class(self) -> type[WeibullEP]:
        return WeibullEP

    @override
    def nat_class(self) -> type[WeibullNP]:
        return WeibullNP


class JointInfo(DistributionInfo[JointDistributionN, JointDistributionE, dict[str, Any]]):
    def __init__(self, infos: Mapping[str, DistributionInfo[Any, Any, Any]]) -> None:
        super().__init__()
        self.infos = dict(infos)

    @override
    def nat_to_scipy_distribution(self, q: JointDistributionN) -> Any:
        return ScipyJointDistribution(
                {name: info.nat_to_scipy_distribution(q.sub_distributions()[name])
                 for name, info in self.infos.items()})

    @override
    def scipy_to_exp_family_observation(self, x: dict[str, Any]) -> dict[str, Any]:
        assert isinstance(x, dict)
        return {name: info.scipy_to_exp_family_observation(x[name])
                for name, info in self.infos.items()}

    @override
    def exp_structure(self) -> Structure[JointDistributionE]:
        infos = []
        for name, info in self.infos.items():
            infos.extend([SubDistributionInfo((*sub_info.path, name),
                                              sub_info.type_,
                                              sub_info.dimensions,
                                              sub_info.sub_distribution_names)
                          for sub_info in info.exp_structure().infos])
        infos.append(SubDistributionInfo((), self.exp_class(), self.dimensions,
                                         list(self.infos.keys())))
        return Structure(infos)

    @override
    def nat_structure(self) -> Structure[JointDistributionN]:
        infos = []
        for name, info in self.infos.items():
            infos.extend([SubDistributionInfo((*sub_info.path, name),
                                              sub_info.type_,
                                              sub_info.dimensions,
                                              sub_info.sub_distribution_names)
                          for sub_info in info.nat_structure().infos])
        infos.append(SubDistributionInfo((), self.nat_class(), self.dimensions,
                                         list(self.infos.keys())))
        return Structure(infos)

    @override
    def exp_class(self) -> type[JointDistributionE]:
        return JointDistributionE

    @override
    def nat_class(self) -> type[JointDistributionN]:
        return JointDistributionN


def create_infos() -> list[DistributionInfo[Any, Any, Any]]:
    return [
            BernoulliInfo(),
            BetaInfo(),
            ChiInfo(),
            ChiSquareInfo(),
            ComplexCircularlySymmetricNormalInfo(dimensions=3),
            ComplexMultivariateUnitNormalInfo(dimensions=4),
            ComplexNormalInfo(),
            ComplexUnitNormalInfo(),
            DirichletInfo(dimensions=5),
            ExponentialInfo(),
            GammaInfo(),
            GeneralizedDirichletInfo(dimensions=5),
            GeometricInfo(),
            IsotropicNormalInfo(dimensions=5),
            LogarithmicInfo(),
            MultivariateDiagonalNormalInfo(dimensions=4),
            MultivariateFixedVarianceNormalInfo(dimensions=2),
            MultivariateNormalInfo(dimensions=4),
            MultivariateUnitNormalInfo(dimensions=5),
            NegativeBinomialInfo(),
            JointInfo(infos={'gamma': GammaInfo(), 'normal': NormalInfo()}),
            NormalInfo(),
            PoissonInfo(),
            RayleighInfo(),
            UnitNormalInfo(),
            VonMisesInfo(),
            VonMisesFisherInfo(dimensions=5),
            WeibullInfo(),
            ]



from __future__ import annotations

import math
from typing import Any

import jax
from tjax import JaxComplexArray, JaxRealArray, KeyArray, Shape, outer_product
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import (SymmetricMatrixSupport, VectorSupport, complex_field,
                          distribution_parameter)


@dataclass
class ComplexCircularlySymmetricNormalNP(
        HasEntropyNP['ComplexCircularlySymmetricNormalEP'],
        NaturalParametrization['ComplexCircularlySymmetricNormalEP', JaxComplexArray],
        Multidimensional,
        Samplable):
    """The natural parameters of the circularly symmetric complex multivariate normal distribution.

    This has zero mean and and zero pseudo-variance. This is a curved exponential family.

    Args:
        negative_precision: -1/Var(x).
    """
    negative_precision: JaxComplexArray = distribution_parameter(
        SymmetricMatrixSupport(hermitian=True, negative_semidefinite=True))
    # S = -1/negative_precision, U = 0
    # P = S.conjugate, R = 0
    # H = -1/S, J = 0
    # K = 0, L = S/2
    # eta = 0
    # Leta = 0

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_precision.shape[:-2]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport(ring=complex_field)

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        log_det_s = xp.log(xp.real(xp.linalg.det(-self.negative_precision)))
        return -log_det_s + self.dimensions() * math.log(math.pi)

    @override
    def to_exp(self) -> ComplexCircularlySymmetricNormalEP:
        xp = self.get_namespace()
        return ComplexCircularlySymmetricNormalEP(
            xp.conj(xp.linalg.inv(-self.negative_precision)))

    @override
    def carrier_measure(self, x: JaxComplexArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxComplexArray, **fixed_parameters: Any
                              ) -> ComplexCircularlySymmetricNormalEP:
        return ComplexCircularlySymmetricNormalEP(outer_product(x, x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        return self.to_exp().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.negative_precision.shape[-1]


@dataclass
class ComplexCircularlySymmetricNormalEP(
        HasEntropyEP[ComplexCircularlySymmetricNormalNP],
        ExpectationParametrization[ComplexCircularlySymmetricNormalNP],
        Multidimensional,
        Samplable):
    """The expectation parameters of the circularly symmetric complex normal distribution.

    This has zero mean and and zero pseudo-variance. This is a curved exponential family.

    Args:
        variance: Var(x).
    """
    variance: JaxComplexArray = distribution_parameter(
            SymmetricMatrixSupport(hermitian=True, positive_semidefinite=True))

    @property
    @override
    def shape(self) -> Shape:
        return self.variance.shape[:-2]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport(ring=complex_field)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ComplexCircularlySymmetricNormalNP]:
        return ComplexCircularlySymmetricNormalNP

    @override
    def to_nat(self) -> ComplexCircularlySymmetricNormalNP:
        xp = self.get_namespace()
        return ComplexCircularlySymmetricNormalNP(xp.conj(-xp.linalg.inv(self.variance)))

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        n = self.dimensions()
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        xy_rvs = jax.random.multivariate_normal(key,
                                                self._multivariate_normal_mean(),
                                                self._multivariate_normal_cov(),
                                                shape)
        return xy_rvs[..., :n] + 1j * xy_rvs[..., n:]

    @override
    def dimensions(self) -> int:
        return self.variance.shape[-1]

    def _multivariate_normal_mean(self) -> JaxRealArray:
        """Return the mean of a corresponding real distribution with double the size."""
        xp = self.get_namespace()
        n = self.dimensions()
        return xp.zeros((*self.shape, n * 2))

    def _multivariate_normal_cov(self) -> JaxRealArray:
        """Return the covariance of a corresponding real distribution with double the size."""
        xp = self.get_namespace()
        gamma_r = 0.5 * xp.real(self.variance)
        gamma_i = 0.5 * xp.imag(self.variance)
        return xp.concat([xp.concat([gamma_r, -gamma_i], axis=-1),
                          xp.concat([gamma_i, gamma_r], axis=-1)],
                         axis=-2)



from __future__ import annotations

from typing import Any

from jax.random import geometric
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter, negative_support, positive_support
from ..parametrization import SimpleDistribution
from .negative_binomial_common import NBCommonEP, NBCommonNP


@dataclass
class GeometricNP(HasEntropyNP['GeometricEP'],
                  Samplable,
                  NBCommonNP['GeometricEP'],
                  NaturalParametrization['GeometricEP', JaxRealArray],
                  SimpleDistribution):
    """The natural parameters of the geometric distribution.

    Models the number of Bernoulli trials having probability p until one failure.  Thus, it has
    support {0, ...}.

    Args:
        log_not_p: log(1-p).
    """
    log_not_p: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))

    @override
    def to_exp(self) -> GeometricEP:
        return GeometricEP(self._mean())

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> GeometricEP:
        return GeometricEP(x)

    @override
    def _failures(self) -> int:
        return 1

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class GeometricEP(HasEntropyEP[GeometricNP],
                  Samplable,
                  NBCommonEP[GeometricNP],
                  ExpectationParametrization[GeometricNP],
                  SimpleDistribution):
    """The expectation parameters of the geometric distribution.

    Models the number of Bernoulli trials having probability p until one failure.  Thus, it has
    support {0, ...}.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[GeometricNP]:
        return GeometricNP

    @override
    def to_nat(self) -> GeometricNP:
        return GeometricNP(self._log_not_p())

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.mean.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        p = 1.0 / self.mean
        return geometric(key, p, shape)

    @override
    def _failures(self) -> int:
        return 1



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
import numpy as np
from numpy.random import Generator
from tjax import assert_tree_allclose

from efax import MultivariateNormalNP

from .create_info import MultivariateNormalInfo
from .distribution_info import DistributionInfo


def test_mvn_fisher_information(distribution_name: str | None) -> None:
    MultivariateNormalInfo.skip_if_deselected(distribution_name)
    m = MultivariateNormalNP(jnp.zeros(2), -0.5 * jnp.eye(2))

    assert_tree_allclose(m._fisher_information_matrix(),  # noqa: SLF001
                         np.asarray([[1., 0., 0., 0., 0.],
                                     [0., 1., 0., 0., 0.],
                                     [0., 0., 2., 0., 0.],
                                     [0., 0., 0., 4., 0.],
                                     [0., 0., 0., 0., 2.]]))

    assert_tree_allclose(
        m.fisher_information_diagonal(),
        MultivariateNormalNP(jnp.asarray([1.0, 1.0]),
                             jnp.asarray([[2., 4.], [4., 2.]])))

    assert_tree_allclose(m.fisher_information_trace(),
                         MultivariateNormalNP(jnp.asarray(2.0), jnp.asarray(8.0)))


def test_mvn_fisher_information_b(distribution_name: str | None) -> None:
    MultivariateNormalInfo.skip_if_deselected(distribution_name)
    m = MultivariateNormalNP(jnp.asarray([3.0, 5.0]), jnp.asarray([[-0.5, 0.0], [0.0, -0.8]]))

    assert_tree_allclose(m._fisher_information_matrix(),  # noqa: SLF001
                         jnp.asarray([[1., -0., 6., 6.25, -0.],
                                      [-0., 0.625, -0., 3.75, 3.90625],
                                      [6., 0., 38., 37.5, 0.],
                                      [6.25, 3.75, 37.5, 64.062, 23.4375],
                                      [0., 3.90625, 0., 23.4375, 25.1953]]))

    assert_tree_allclose(
        m.fisher_information_diagonal(),
        MultivariateNormalNP(jnp.asarray([1.0, 0.625]),
                             jnp.asarray([[38., 64.062], [64.062, 25.1953]])))

    assert_tree_allclose(m.fisher_information_trace(),
                         MultivariateNormalNP(jnp.asarray(1.625), jnp.asarray(127.258)))


def test_fisher_information_is_convex(generator: Generator,
                                      distribution_info: DistributionInfo[Any, Any, Any]) -> None:
    shape = (3, 2)
    nat_parameters = distribution_info.nat_parameter_generator(generator, shape=shape)
    fisher_information = nat_parameters._fisher_information_matrix()  # noqa: SLF001
    assert issubclass(fisher_information.dtype.type, jnp.floating)
    eigvals = jnp.linalg.eigvals(fisher_information)
    if not jnp.all(eigvals >= 0.0):
        msg = (f"The Fisher information of {nat_parameters} is not convex.  Its eigenvalues are:"
               f"{eigvals}")
        raise AssertionError(msg)
    determinant = jnp.linalg.det(fisher_information)
    assert determinant.shape == shape
    assert issubclass(determinant.dtype.type, jnp.floating)

    if not jnp.all(determinant >= 0.0):
        msg = (f"The determinant of the Fisher information of {nat_parameters} is not all "
               f"nonnegative: {determinant}")
        raise AssertionError(msg)



"""Optimization.

This example illustrates how this library fits in a typical machine learning context.  Suppose we
have an unknown target value, and a loss function based on the cross-entropy between the target
value and a predictive distribution.  We will optimize the predictive distribution by a small
fraction of its cotangent.
"""
import jax.numpy as jnp
from jax import grad, lax
from tjax import JaxBooleanArray, JaxRealArray, jit, print_generic

from efax import BernoulliEP, BernoulliNP, parameter_dot_product, parameter_map


def cross_entropy_loss(p: BernoulliEP, q: BernoulliNP) -> JaxRealArray:
    return jnp.sum(p.cross_entropy(q))


gradient_cross_entropy = jit(grad(cross_entropy_loss, 1))


def apply(x: JaxRealArray, x_bar: JaxRealArray) -> JaxRealArray:
    return x - 1e-4 * x_bar


def body_fun(q: BernoulliNP) -> BernoulliNP:
    q_bar = gradient_cross_entropy(target_distribution, q)
    return parameter_map(apply, q, q_bar)


def cond_fun(q: BernoulliNP) -> JaxBooleanArray:
    q_bar = gradient_cross_entropy(target_distribution, q)
    total = jnp.sum(parameter_dot_product(q_bar, q_bar))
    return total > 1e-6  # noqa: PLR2004


# The target_distribution is represented as the expectation parameters of a Bernoulli distribution
# corresponding to probabilities 0.3, 0.4, and 0.7.
target_distribution = BernoulliEP(jnp.asarray([0.3, 0.4, 0.7]))

# The initial predictive distribution is represented as the natural parameters of a Bernoulli
# distribution corresponding to log-odds 0, which is probability 0.5.
initial_predictive_distribution = BernoulliNP(jnp.zeros(3))

# Optimize the predictive distribution iteratively, and output the natural parameters of the
# prediction.
predictive_distribution = lax.while_loop(cond_fun, body_fun, initial_predictive_distribution)
print_generic(predictive_distribution)
# BernoulliNP
# └── log_odds=Jax Array (3,) float32
#     └──  -0.8440 │ -0.4047 │ 0.8440

# Compare the optimized predictive distribution with the target value in the same parametrization.
print_generic(target_distribution.to_nat())
# BernoulliNP
# └── log_odds=Jax Array (3,) float32
#     └──  -0.8473 │ -0.4055 │ 0.8473

# Print the optimized natural parameters as expectation parameters.
print_generic(predictive_distribution.to_exp())
# BernoulliEP
# └── probability=Jax Array (3,) float32
#     └──  0.3007 │ 0.4002 │ 0.6993



from __future__ import annotations

import math
from typing import Any, Self

import jax
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.conjugate_prior import HasConjugatePrior
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import ScalarSupport, distribution_parameter
from .normal import NormalNP


@dataclass
class UnitNormalNP(HasEntropyNP['UnitNormalEP'],
                   NaturalParametrization['UnitNormalEP', JaxRealArray],
                   Samplable):
    """The natural parametrization of the normal distribution with unit variance.

    This is a curved exponential family.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return 0.5 * (xp.square(self.mean) + math.log(math.pi * 2.0))

    @override
    def to_exp(self) -> UnitNormalEP:
        return UnitNormalEP(self.mean)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        # The second moment of a delta distribution at x.
        xp = self.get_namespace(x)
        return -0.5 * xp.square(x)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray,
                              **fixed_parameters: Any
                              ) -> UnitNormalEP:
        return UnitNormalEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        return jax.random.normal(key, shape) + self.mean


@dataclass
class UnitNormalEP(HasEntropyEP[UnitNormalNP],
                   HasConjugatePrior,
                   Samplable,
                   ExpectationParametrization[UnitNormalNP]):
    """The expectation parametrization of the normal distribution with unit variance.

    This is a curved exponential family.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[UnitNormalNP]:
        return UnitNormalNP

    @override
    def to_nat(self) -> UnitNormalNP:
        return UnitNormalNP(self.mean)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        # The second moment of a normal distribution with the given mean.
        xp = self.get_namespace()
        return -0.5 * (xp.square(self.mean) + 1.0)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> NormalNP:
        negative_half_precision = -0.5 * n
        return NormalNP(n * self.mean, negative_half_precision)

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, NormalNP)
        n = -2.0 * cp.negative_half_precision
        mean = cp.mean_times_precision / n
        return (cls(mean), n)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean



from __future__ import annotations

from typing import Any

import jax
import numpy as np
from jax.nn import one_hot
from jax.scipy import special as jss
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..array_tools import array_at
from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasGeneralizedConjugatePrior
from ..interfaces.multidimensional import Multidimensional
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, VectorSupport, distribution_parameter
from .dirichlet import DirichletNP
from .gen_dirichlet import GeneralizedDirichletNP


@dataclass
class MultinomialNP(HasEntropyNP['MultinomialEP'],
                    Samplable,
                    NaturalParametrization['MultinomialEP', JaxRealArray], Multidimensional):
    """The natural parametrization of the multinomial distribution.

    Args:
        log_odds: [log(p_i / p_n)]_{i in 1...n-1}.
    """
    log_odds: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_odds.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        max_q = xp.maximum(0.0, xp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_a = xp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        return max_q + log_scaled_a

    @override
    def to_exp(self) -> MultinomialEP:
        xp = self.get_namespace()
        max_q = xp.maximum(0.0, xp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_a = xp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        return MultinomialEP(xp.exp(q_minus_max_q - log_scaled_a[..., np.newaxis]))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultinomialEP:
        return MultinomialEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        return one_hot(jax.random.categorical(key, self.log_odds, shape=shape),
                       self.dimensions())

    @override
    def dimensions(self) -> int:
        return self.log_odds.shape[-1]

    def nat_to_probability(self) -> JaxRealArray:
        xp = self.get_namespace()
        max_q = xp.maximum(0.0, xp.amax(self.log_odds, axis=-1))
        q_minus_max_q = self.log_odds - max_q[..., np.newaxis]
        log_scaled_a = xp.logaddexp(-max_q, jss.logsumexp(q_minus_max_q, axis=-1))
        p = xp.exp(q_minus_max_q - log_scaled_a[..., np.newaxis])
        final_p = 1.0 - xp.sum(p, axis=-1, keepdims=True)
        return xp.concat((p, final_p), axis=-1)

    def nat_to_surprisal(self) -> JaxRealArray:
        xp = self.get_namespace()
        total_p = self.nat_to_probability()
        return -xp.log(total_p)


@dataclass
class MultinomialEP(HasEntropyEP[MultinomialNP],
                    Samplable,
                    HasGeneralizedConjugatePrior,
                    ExpectationParametrization[MultinomialNP],
                    Multidimensional):
    """The expectation parametrization of the multinomial distribution.

    Args:
        log_odds: The probability vector with the final element omitted, i.e., [p_i]_{i in 1...n-1}.
    """
    probability: JaxRealArray = distribution_parameter(VectorSupport(ring=RealField(
        minimum=0.0, maximum=1.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.probability.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultinomialNP]:
        return MultinomialNP

    @override
    def to_nat(self) -> MultinomialNP:
        xp = self.get_namespace()
        p_k = 1.0 - xp.sum(self.probability, axis=-1, keepdims=True)
        return MultinomialNP(xp.log(self.probability / p_k))

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> DirichletNP:
        xp = self.get_namespace()
        reshaped_n = n[..., np.newaxis]
        final_p = 1.0 - xp.sum(self.probability, axis=-1, keepdims=True)
        return DirichletNP(reshaped_n * xp.concat((self.probability, final_p), axis=-1))

    @override
    def generalized_conjugate_prior_distribution(self, n: JaxRealArray) -> GeneralizedDirichletNP:
        xp = self.get_namespace()
        final_p = 1.0 - xp.sum(self.probability, axis=-1, keepdims=True)
        all_p = xp.concat((self.probability, final_p), axis=-1)
        alpha = n * all_p
        beta = n * (1.0 - all_p)
        alpha_roll = array_at(xp.roll(alpha, -1, axis=-1))[..., -1].set(0.0)
        gamma = -xp.diff(beta, append=1.0) - alpha_roll
        return GeneralizedDirichletNP(alpha - 1.0, gamma)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.probability

    @override
    def dimensions(self) -> int:
        return self.probability.shape[-1]

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from jax import grad, vmap
from numpy.random import Generator
from numpy.testing import assert_allclose
from tjax import assert_tree_allclose

from efax import HasConjugatePrior, HasGeneralizedConjugatePrior, MaximumLikelihoodEstimator

from .distribution_info import DistributionInfo


def test_conjugate_prior(generator: Generator,
                         cp_distribution_info: DistributionInfo[Any, Any, Any],
                         distribution_name: str | None) -> None:
    """Test that the conjugate prior actually matches the distribution."""
    cp_distribution_info.skip_if_deselected(distribution_name)
    shape = (4, 3)
    n = 100.0 * jnp.ones(shape)

    # Choose a random distribution.
    p = cp_distribution_info.exp_parameter_generator(generator, shape=shape)
    assert isinstance(p, HasConjugatePrior)

    # Find its conjugate prior at that point with many observations.
    cp_q = p.conjugate_prior_distribution(n)

    assert cp_q.shape == p.shape

    # Produce a copy of p that matches the conjugate prior distribution.
    cp_x = p.conjugate_prior_observation()

    # Produce a function that calculates the gradient of the density with respect to p.  Ensure that
    # it is broadcasted according to the shape.
    density_gradient = grad(type(cp_q).pdf, argnums=1)
    for _ in range(len(shape)):
        density_gradient = vmap(density_gradient)

    # Check the gradient of the density of the conjugate prior at p is zero.
    derivative = density_gradient(cp_q, cp_x)
    zero_derivative = jnp.zeros_like(derivative)
    assert_tree_allclose(derivative, zero_derivative, atol=1.5)


def test_from_conjugate_prior(generator: Generator,
                              cp_distribution_info: DistributionInfo[Any, Any, Any],
                              distribution_name: str | None) -> None:
    """Test that the conjugate prior is reversible."""
    cp_distribution_info.skip_if_deselected(distribution_name)
    shape = (4, 3)
    n = 123.0 * jnp.ones(shape)

    # Choose a random distribution.
    p = cp_distribution_info.exp_parameter_generator(generator, shape=shape)
    assert isinstance(p, HasConjugatePrior)

    # Find its conjugate prior at that point with many observations.
    cp_q = p.conjugate_prior_distribution(n)
    p_estimator = MaximumLikelihoodEstimator.create_estimator(p)
    p_prime, n_prime = p_estimator.from_conjugate_prior_distribution(cp_q)
    assert_tree_allclose(p, p_prime, atol=1.5)
    assert_allclose(n, n_prime, atol=1.5)


def test_generalized_conjugate_prior(generator: Generator,
                                     gcp_distribution_info: DistributionInfo[Any, Any, Any],
                                     distribution_name: str | None
                                     ) -> None:
    """Same as test_conjugate_prior, but with generalized_conjugate_prior_distribution."""
    gcp_distribution_info.skip_if_deselected(distribution_name)
    shape = (4, 3)

    # Choose a random distribution.
    p = gcp_distribution_info.exp_parameter_generator(generator, shape=shape)
    assert isinstance(p, HasGeneralizedConjugatePrior)

    # Find its conjugate prior at that point with many observations.
    n = 100.0 * jnp.ones((*shape, p.dimensions()))
    gcp_q = p.generalized_conjugate_prior_distribution(n)

    assert gcp_q.shape == p.shape

    # Produce a copy of p that matches the conjugate prior distribution.
    cp_x = p.conjugate_prior_observation()

    # Produce a function that calculates the gradient of the density with respect to p.  Ensure that
    # it is broadcasted according to the shape.
    density_gradient = grad(type(gcp_q).pdf, argnums=1)
    for _ in range(len(shape)):
        density_gradient = vmap(density_gradient)

    # Check the gradient of the density of the conjugate prior at p is zero.
    derivative = density_gradient(gcp_q, cp_x)
    zero_derivative = jnp.zeros_like(derivative)
    assert_tree_allclose(derivative, zero_derivative, atol=1.5)



from __future__ import annotations

from typing import Any

import numpy as np
import pytest
from numpy.random import Generator as NumpyGenerator
from tjax import Shape

from efax import ScipyDirichlet, ScipyMultivariateNormal


@pytest.mark.nondistribution
@pytest.mark.parametrize(('distribution', 'm', 'n'),
                         [(ScipyMultivariateNormal.from_mc(mean=np.zeros((3, 4))), (3,), 4),
                          (ScipyMultivariateNormal.from_mc(cov=np.eye(2)), (), 2),
                          (ScipyMultivariateNormal.from_mc(), (), 1),
                          (ScipyDirichlet(alpha=np.ones((2, 5))), (2,), 5),
                          (ScipyDirichlet(alpha=np.ones((2, 3, 5))), (2, 3), 5),
                          ])
def test_shaped(generator: NumpyGenerator, distribution: Any, m: Shape, n: int) -> None:
    assert distribution.rvs().shape == (*m, n)
    assert distribution.rvs(1).shape == (*m, 1, n)
    assert distribution.rvs((1,)).shape == (*m, 1, n)
    assert distribution.rvs((1, 1, 1)).shape == (*m, 1, 1, 1, n)
    assert distribution.rvs((1, 2, 3)).shape == (*m, 1, 2, 3, n)



from collections.abc import Iterable, Mapping
from dataclasses import fields
from typing import Any, Literal, overload

from tjax import JaxComplexArray

from .parameter import Support
from .parametrization import Distribution, SimpleDistribution
from .types import Path


def flatten_mapping(m: Mapping[str, Any], /) -> dict[Path, Any]:
    """Flatten a nested mapping."""
    result: dict[Path, Any] = {}

    def _flatten(m: Mapping[str, Mapping[str, Any] | Any], prefix: Path) -> None:
        for key, value in m.items():
            path = (*prefix, key)
            if isinstance(value, Mapping):
                _flatten(value, path)
                continue
            result[path] = value
    _flatten(m, ())

    return result


def unflatten_mapping(m: Mapping[Path, Any], /) -> dict[str, Any]:
    """Unflatten a nested mapping."""
    result: dict[str, Any] = {}
    for path, value in m.items():
        cursor = result
        for key in path[:-1]:
            cursor = cursor.setdefault(key, {})
        cursor[path[-1]] = value
    return result


# def parameters(p: Distribution,
#                     /,
#                     *,
#                     fixed: bool | None = None
#                     ) -> dict[str, dict[str, Any] | tuple[JaxComplexArray, Support]]:
#     """Walk the parameters of a distribution and return the variable ones.
#
#     Args:
#         p: The parametrization to walk.
#         fixed: If true or false, return the fixed or variable parameters, otherwise return both.
#
#     Returns:
#         The path, value, and support of each variable parameter.
#     """
#     return ({name: (value, support)
#              for name, value, support, is_fixed in p.parameters()
#              if fixed is None or is_fixed == fixed}
#             | {name: parameters(value, fixed=fixed)
#                 for name, value in p.sub_distributions()})


@overload
def parameters(p: Distribution,
               /,
               *,
               fixed: bool | None = None,
               support: Literal[False] = False,
               recurse: Literal[False],
               ) -> dict[str, JaxComplexArray]:
    ...


@overload
def parameters(p: Distribution,
               /,
               *,
               fixed: bool | None = None,
               support: Literal[True],
               recurse: Literal[False],
               ) -> dict[str, tuple[JaxComplexArray, Support]]:
    ...


@overload
def parameters(p: Distribution,
               /,
               *,
               fixed: bool | None = None,
               support: Literal[False] = False,
               recurse: Literal[True] = True,
               ) -> dict[Path, JaxComplexArray]:
    ...


@overload
def parameters(p: Distribution,
               /,
               *,
               fixed: bool | None = None,
               support: Literal[True],
               recurse: Literal[True] = True,
               ) -> dict[Path, tuple[JaxComplexArray, Support]]:
    ...


def parameters(p: Distribution,
               /,
               *,
               fixed: bool | None = None,
               support: bool = False,
               recurse: bool = True,
               ) -> Any:
    """Return the parameters of a distribution.

    Args:
        p: The parametrization to walk.
        fixed: If true or false, return the fixed or variable parameters, otherwise return both.
        support: If true, yield the support also.
        recurse: If true, recurse into sub-distributions.

    Returns:
        The path, value, and support of each variable parameter.
    """
    def _parameters(q: Distribution,
                    base_path: Path
                    ) -> Iterable[tuple[Any, ...]]:
        for this_field in fields(q):
            name = this_field.name
            value = getattr(q, name)
            metadata = this_field.metadata
            if not metadata.get('parameter', False):
                continue
            is_fixed = metadata['fixed']
            if not isinstance(is_fixed, bool):
                raise TypeError
            if fixed is not None and is_fixed != fixed:
                continue
            this_path = (*base_path, name) if recurse else name
            if support:
                yield this_path, value, metadata['support']
            else:
                yield this_path, value
        if not recurse:
            return
        for name, value in q.sub_distributions().items():
            this_path = (*base_path, name)
            yield from _parameters(value, this_path)
    if support:
        return {key: (value, support) for key, value, support in _parameters(p, ())}
    return dict(_parameters(p, ()))


def support(p: type[Distribution] | Distribution,
            /,
            *,
            fixed: bool | None = None,
            ) -> dict[str, Support]:
    """Return the support of a distribution class.

    Args:
        p: The parametrization to walk.
        fixed: If true or false, return the fixed or variable parameters, otherwise return both.

    Returns:
        The path, value, and support of each variable parameter.
    """
    def _parameters(q: type[Distribution],
                    base_path: Path
                    ) -> Iterable[tuple[str, Support]]:
        for this_field in fields(q):
            name = this_field.name
            metadata = this_field.metadata
            if not metadata.get('parameter', False):
                continue
            support = metadata['support']
            is_fixed = metadata['fixed']
            if not isinstance(is_fixed, bool):
                raise TypeError
            if fixed is not None and is_fixed != fixed:
                continue
            if not isinstance(support, Support):
                raise TypeError
            yield name, support
    cls_p: type[Distribution] = type(p) if isinstance(p, Distribution) else p
    return dict(_parameters(cls_p, ()))


def flat_dict_of_parameters(d: Distribution) -> dict[Path, SimpleDistribution]:
    from .transform.joint import JointDistribution  # noqa: PLC0415
    if isinstance(d, JointDistribution):
        return flatten_mapping(d.as_dict())
    assert isinstance(d, SimpleDistribution)
    return {(): d}


def flat_dict_of_observations(x: Mapping[str, Any] | JaxComplexArray
                              ) -> dict[Path, JaxComplexArray]:
    return flatten_mapping(x) if isinstance(x, Mapping) else {(): x}



from __future__ import annotations

from dataclasses import KW_ONLY, field
from typing import Any, Generic, Self, TypeAlias, TypeVar

from jax import vmap
from tjax import JaxRealArray, jit
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...natural_parametrization import NaturalParametrization
from ...parametrization import SimpleDistribution
from ...structure import Flattener

NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])
SP: TypeAlias = JaxRealArray


class ExpToNatMinimizer:
    def solve(self, exp_to_nat: ExpToNat[Any]) -> SP:
        raise NotImplementedError


@dataclass
class ExpToNat(ExpectationParametrization[NP], SimpleDistribution, Generic[NP]):
    """This mixin implements the conversion from expectation to natural parameters.

    It uses Newton's method with a Jacobian to invert the gradient log-normalizer.
    """
    _: KW_ONLY
    minimizer: ExpToNatMinimizer | None = field(default=None, repr=False)

    def __post_init__(self) -> None:
        if hasattr(super(), '__post_init__'):
            super().__post_init__()  # type: ignore # pyright: ignore
        if self.minimizer is None:
            from .optimistix import default_bisection_minimizer, default_minimizer  # noqa: PLC0415
            initial_search_parameters = self.initial_search_parameters()
            object.__setattr__(self,
                               'minimizer',
                               default_minimizer
                               if initial_search_parameters.shape[-1] > 1
                               else default_bisection_minimizer)

    @jit
    @override
    def to_nat(self) -> NP:
        flattener, flattened = Flattener[Self].flatten(self)  # type: ignore[arg-type]

        def solve(flattener: Flattener[Self], flattened: JaxRealArray) -> JaxRealArray:
            x = flattener.unflatten(flattened)
            assert self.minimizer is not None
            return self.minimizer.solve(x)

        for _ in range(self.ndim):
            solve = vmap(solve, in_axes=(None, 0))
        return self.search_to_natural(solve(flattener, flattened))

    def initial_search_parameters(self) -> SP:
        """The initial value of the parameters used by the search algorithm.

        Returns a real array of shape (self.shape, n) for some n.
        """
        _, flattened = Flattener.flatten(self)
        xp = self.get_namespace()
        return xp.zeros_like(flattened)

    def search_to_natural(self, search_parameters: SP) -> NP:
        """Convert the search parameters to the natural parametrization.

        Args:
            search_parameters: The parameters in search space.
        Returns: The corresponding natural parameters.

        This function should be monotonic.
        """
        np_cls = self.natural_parametrization_cls()
        assert issubclass(np_cls, SimpleDistribution)
        flattener = Flattener.create_flattener(self,
                                               override_unflattened_type=np_cls,
                                               mapped_to_plane=True)
        return flattener.unflatten(search_parameters)

    def search_gradient(self, search_parameters: SP) -> SP:
        """Convert the search parameters to the natural gradient.

        This function converts natural parameters to expectations parameters, takes the difference
        with self, and flattens the result.

        Args:
            search_parameters: The parameters in search space.
        Returns: The gradient of the loss wrt to the search parameters.
        """
        search_np = self.search_to_natural(search_parameters)
        search_ep: ExpToNat[NP] = search_np.to_exp()
        _, self_flat = Flattener.flatten(self, map_to_plane=False)
        _, search_flat = Flattener.flatten(search_ep, map_to_plane=False)
        return search_flat - self_flat



from __future__ import annotations

from typing import Self

import numpy as np
from numpy.random import Generator
from tjax import NumpyComplexArray, NumpyComplexNumeric, NumpyRealArray, NumpyRealNumeric, ShapeLike
from typing_extensions import override

from .multivariate_normal import ScipyMultivariateNormal, ScipyMultivariateNormalUnvectorized
from .shaped_distribution import ShapedDistribution


class ScipyComplexNormalUnvectorized:
    """Represents an array of univariate complex normal distributions."""
    @override
    def __init__(self,
                 mean: NumpyComplexNumeric,
                 variance: NumpyRealNumeric,
                 pseudo_variance: NumpyComplexNumeric) -> None:
        super().__init__()
        self.mean: NumpyComplexArray = np.asarray(mean)
        self.variance: NumpyRealArray = np.asarray(variance)
        self.pseudo_variance: NumpyComplexArray = np.asarray(pseudo_variance)
        if not np.issubdtype(self.variance.dtype, np.floating):
            msg = f"The variance {self.variance} has non-real dtype {self.variance.dtype}."
            raise TypeError(msg)
        if np.any(np.abs(self.pseudo_variance) > self.variance):
            msg = f"The pseudo-variance {pseudo_variance} is bigger than the variance {variance}."
            raise ValueError(msg)
        if not (self.mean.shape == self.variance.shape
                == self.pseudo_variance.shape):
            msg = "Shape mismatch."
            raise ValueError(msg)

    def pdf(self, z: NumpyComplexNumeric, out: None = None) -> NumpyRealNumeric:
        zr = np.stack([np.real(z), np.imag(z)], axis=-1)
        return self.as_multivariate_normal().pdf(zr)

    def rvs(self, size: ShapeLike = (), random_state: Generator | None = None) -> NumpyComplexArray:
        if isinstance(size, int):
            size = (size,)
        if random_state is None:
            random_state = np.random.default_rng()
        xy_rvs = random_state.multivariate_normal(mean=self._multivariate_normal_mean(),
                                                  cov=self._multivariate_normal_cov(),
                                                  size=size)
        return xy_rvs[..., 0] + 1j * xy_rvs[..., 1]

    def entropy(self) -> float:
        return self.as_multivariate_normal().entropy().item()

    def as_multivariate_normal(self) -> ScipyMultivariateNormalUnvectorized:
        mv_mean = self._multivariate_normal_mean()
        mv_cov = self._multivariate_normal_cov()
        return ScipyMultivariateNormalUnvectorized(mean=mv_mean, cov=mv_cov)

    def _multivariate_normal_mean(self) -> NumpyRealArray:
        return np.stack([np.real(self.mean), np.imag(self.mean)], axis=-1)

    def _multivariate_normal_cov(self) -> NumpyRealArray:
        xx = self.variance + np.real(self.pseudo_variance)
        xy = np.imag(self.pseudo_variance)
        yy = self.variance - np.real(self.pseudo_variance)
        xx_xy = np.stack([xx, xy], axis=-1)
        yx_yy = np.stack([xy, yy], axis=-1)
        return 0.5 * np.stack([xx_xy, yx_yy], axis=-2)


class ScipyComplexNormal(ShapedDistribution[ScipyComplexNormalUnvectorized]):
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self,
                 mean: NumpyComplexArray | None = None,
                 variance: NumpyRealArray | None = None,
                 pseudo_variance: NumpyComplexArray | None = None) -> None:
        if mean is not None:
            shape = mean.shape
        elif variance is not None:
            shape = variance.shape
        elif pseudo_variance is not None:
            shape = pseudo_variance.shape
        else:
            raise ValueError
        dtype = np.result_type(*[x.dtype
                                 for x in (mean, variance, pseudo_variance)
                                 if x is not None])
        rvs_shape = ()
        if mean is None:
            mean = np.zeros(shape, dtype=dtype)
        if variance is None:
            variance = np.real(np.ones(shape, dtype=dtype))
        if pseudo_variance is None:
            pseudo_variance = np.zeros(shape, dtype=dtype)

        objects = np.empty(shape, dtype=ScipyComplexNormalUnvectorized)
        for i in np.ndindex(*shape):
            objects[i] = ScipyComplexNormalUnvectorized(mean[i], variance[i], pseudo_variance[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    @classmethod
    def init_using_angle(cls,
                         mean: NumpyComplexNumeric,
                         variance: NumpyRealNumeric,
                         angle: NumpyRealNumeric,
                         polarization: NumpyComplexNumeric) -> Self:
        polarization_np: NumpyComplexArray = np.asarray(polarization)
        r: NumpyComplexArray = polarization_np * np.exp(1j * 2 * np.pi * angle * 2)
        mean_array = np.asarray(mean)
        variance_array = np.asarray(variance)
        pseudo_variance_array = np.asarray(r * variance)
        return cls(mean_array, variance_array, pseudo_variance_array)

    def as_multivariate_normal(self) -> ScipyMultivariateNormal:
        objects = np.empty(self.shape, dtype=ScipyMultivariateNormalUnvectorized)
        for i in np.ndindex(*self.shape):
            objects[i] = self.objects[i].as_multivariate_normal()
        return ScipyMultivariateNormal(self.shape, self.rvs_shape, self.real_dtype, objects)

    @property
    def mean(self) -> NumpyComplexArray:
        retval = np.empty(self.shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].mean
        return retval

    @property
    def variance(self) -> NumpyComplexArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].variance
        return retval

    @property
    def pseudo_variance(self) -> NumpyComplexArray:
        retval = np.empty(self.shape, dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].pseudo_variance
        return retval



from __future__ import annotations

import math
from typing import Any, Self

import jax
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.conjugate_prior import HasGeneralizedConjugatePrior
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter
from .diagonal import MultivariateDiagonalNormalNP
from .isotropic import IsotropicNormalNP


@dataclass
class MultivariateUnitNormalNP(HasEntropyNP['MultivariateUnitNormalEP'],
                               NaturalParametrization['MultivariateUnitNormalEP', JaxRealArray],
                               Multidimensional,
                               Samplable):
    """The natural parametrization of the multivariate normal distribution with unit variance.

    This is a curved exponential family.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return 0.5 * (xp.sum(xp.square(self.mean), axis=-1)
                      + self.dimensions() * math.log(math.pi * 2.0))

    @override
    def to_exp(self) -> MultivariateUnitNormalEP:
        return MultivariateUnitNormalEP(self.mean)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        # The second moment of a delta distribution at x.
        xp = self.get_namespace(x)
        return -0.5 * xp.sum(xp.square(x), axis=-1)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateUnitNormalEP:
        return MultivariateUnitNormalEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        return jax.random.normal(key, shape) + self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]


@dataclass
class MultivariateUnitNormalEP(
        HasEntropyEP[MultivariateUnitNormalNP],
        HasGeneralizedConjugatePrior,
        Samplable,
        ExpectationParametrization[MultivariateUnitNormalNP],
        Multidimensional):
    """The expectation parametrization of the multivariate normal distribution with unit variance.

    This is a curved exponential family.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateUnitNormalNP]:
        return MultivariateUnitNormalNP

    @override
    def to_nat(self) -> MultivariateUnitNormalNP:
        return MultivariateUnitNormalNP(self.mean)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        # The second moment of a normal distribution with the given mean.
        xp = self.get_namespace()
        return -0.5 * (xp.sum(xp.square(self.mean), axis=-1) + self.dimensions())

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> IsotropicNormalNP:
        xp = self.get_namespace()
        negative_half_precision = -0.5 * n
        return IsotropicNormalNP(n[..., xp.newaxis] * self.mean, negative_half_precision)

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, IsotropicNormalNP)
        xp = cp.get_namespace()
        n = -2.0 * cp.negative_half_precision
        mean = cp.mean_times_precision / n[..., xp.newaxis]
        return (cls(mean), n)

    @override
    def generalized_conjugate_prior_distribution(self, n: JaxRealArray
                                                 ) -> MultivariateDiagonalNormalNP:
        negative_half_precision = -0.5 * n
        return MultivariateDiagonalNormalNP(n * self.mean, negative_half_precision)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]



"""These tests apply to only samplable distributions."""
from __future__ import annotations

from typing import Any, TypeAlias

import jax.numpy as jnp
from numpy.random import Generator
from tjax import JaxComplexArray, KeyArray, Shape, assert_tree_allclose

from efax import (Distribution, ExpectationParametrization, JointDistribution,
                  MaximumLikelihoodEstimator, NaturalParametrization, Samplable, SimpleDistribution,
                  Structure, flat_dict_of_observations, flatten_mapping, parameter_mean)

from .create_info import (BetaInfo, ComplexCircularlySymmetricNormalInfo,
                          ComplexMultivariateUnitNormalInfo, ComplexNormalInfo, DirichletInfo,
                          GammaInfo, IsotropicNormalInfo, JointInfo, MultivariateDiagonalNormalInfo,
                          MultivariateFixedVarianceNormalInfo, MultivariateNormalInfo,
                          MultivariateUnitNormalInfo, PoissonInfo)
from .distribution_info import DistributionInfo

Path: TypeAlias = tuple[str, ...]


def produce_samples(generator: Generator,
                    key: KeyArray,
                    sampling_distribution_info: DistributionInfo[NaturalParametrization[Any, Any],
                                                                 ExpectationParametrization[Any],
                                                                 Any],
                    distribution_shape: Shape,
                    sample_shape: Shape,
                    *,
                    natural: bool) -> tuple[ExpectationParametrization[Any],
                                            dict[str, Any] | JaxComplexArray]:
    sampling_object: Distribution
    if natural:
        sampling_object = nat_parameters = sampling_distribution_info.nat_parameter_generator(
                generator, distribution_shape)
        exp_parameters = nat_parameters.to_exp()
    else:
        sampling_object = exp_parameters = sampling_distribution_info.exp_parameter_generator(
                generator, distribution_shape)

    if isinstance(sampling_object, Samplable):
        samples = sampling_object.sample(key, sample_shape)
        samples = sampling_object.domain_support().clamp(samples)
    else:
        assert isinstance(sampling_object, JointDistribution)

        def f(x: Distribution, /) -> JaxComplexArray:
            assert isinstance(x, Samplable)
            sample = x.sample(key, sample_shape)
            return x.domain_support().clamp(sample)

        samples = sampling_object.general_method(f)
    return exp_parameters, samples


def verify_sample_shape(distribution_shape: Shape,
                        sample_shape: Shape,
                        structure: Structure[ExpectationParametrization[Any]],
                        flat_map_of_samples: dict[Path, Any]
                        ) -> None:
    ideal_samples_shape = {info.path: (*sample_shape, *distribution_shape,
                                       *info.type_.domain_support().shape(info.dimensions))
                           for info in structure.infos
                           if issubclass(info.type_, SimpleDistribution)}
    samples_shape = {path: s.shape for path, s in flat_map_of_samples.items()}
    assert samples_shape == ideal_samples_shape


def verify_maximum_likelihood_estimate(
        sampling_distribution_info: DistributionInfo[NaturalParametrization[Any, Any],
                                                     ExpectationParametrization[Any],
                                                     Any],
        sample_shape: Shape,
        structure: Structure[ExpectationParametrization[Any]],
        exp_parameters: ExpectationParametrization[Any],
        samples: dict[str, Any] | JaxComplexArray
        ) -> None:
    atol = (1e-1
            if isinstance(sampling_distribution_info,
                          ComplexCircularlySymmetricNormalInfo
                          | MultivariateFixedVarianceNormalInfo)
            else 1e-2
            if isinstance(sampling_distribution_info,
                          ComplexNormalInfo | ComplexMultivariateUnitNormalInfo
                          | MultivariateDiagonalNormalInfo | MultivariateNormalInfo
                          | MultivariateUnitNormalInfo | IsotropicNormalInfo | PoissonInfo
                          | JointInfo)
            else 1e-3
            if isinstance(sampling_distribution_info, JointInfo)
            else 1e-6)
    rtol = (5e-2
            if isinstance(sampling_distribution_info, GammaInfo | DirichletInfo | BetaInfo)
            else 4e-2)
    sample_axes = tuple(range(len(sample_shape)))
    newaxes = (jnp.newaxis,) * len(sample_shape)
    estimator = MaximumLikelihoodEstimator.create_estimator(exp_parameters[*newaxes, ...])
    sampled_exp_parameters = estimator.sufficient_statistics(samples)
    maximum_likelihood_parameters = parameter_mean(sampled_exp_parameters, axis=sample_axes)
    assert_tree_allclose(maximum_likelihood_parameters, exp_parameters, rtol=rtol,
                         atol=atol)


def test_sampling_and_estimation(generator: Generator,
                                 key: KeyArray,
                                 sampling_distribution_info: DistributionInfo[Any, Any, Any],
                                 *,
                                 distribution_name: str | None,
                                 natural: bool) -> None:
    """Test that sampling is consistent with maximum likelihood estimation.

    This tests samples variates from either natural or expectation parametrizations.  Calculates the
    mean of the sufficient statistics, and verifies that it equals the expectation parameters.
    """
    sampling_distribution_info.skip_if_deselected(distribution_name)
    distribution_shape = (4,)  # The number of distributions that are being estimated.
    sample_shape = (1024, 64)  # The number of samples that are taken to do the estimation.
    exp_parameters, samples = produce_samples(generator, key, sampling_distribution_info,
                                              distribution_shape, sample_shape, natural=natural)
    flat_map_of_samples = flat_dict_of_observations(samples)
    structure = Structure.create(exp_parameters)
    flat_map_of_samples = flatten_mapping(samples) if isinstance(samples, dict) else {(): samples}
    verify_sample_shape(distribution_shape, sample_shape, structure, flat_map_of_samples)
    verify_maximum_likelihood_estimate(sampling_distribution_info, sample_shape, structure,
                                       exp_parameters, samples)



from __future__ import annotations

import numpy as np
import scipy.stats as ss
from tjax import NumpyRealArray
from typing_extensions import override

from .shaped_distribution import ShapedDistribution


class ScipyVonMises(ShapedDistribution[ss.vonmises]):
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self, kappa: NumpyRealArray, loc: NumpyRealArray | None = None) -> None:
        if loc is None:
            loc = np.zeros_like(kappa)
        shape = np.broadcast(kappa, loc).shape
        rvs_shape = ()
        dtype = np.result_type(kappa.dtype, loc.dtype)
        kappa = np.astype(np.broadcast_to(kappa, shape), dtype)
        loc = np.astype(np.broadcast_to(loc, shape), dtype)
        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ss.vonmises(kappa[i], loc[i])
        super().__init__(shape, rvs_shape, dtype, objects)


class ScipyVonMisesFisher(ShapedDistribution[ss.vonmises_fisher]):
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self, mu: NumpyRealArray, kappa: NumpyRealArray) -> None:
        assert mu.shape[:-1] == kappa.shape
        shape = kappa.shape
        rvs_shape = (mu.shape[-1],)
        dtype = np.result_type(kappa.dtype, mu.dtype)
        kappa = np.astype(kappa, dtype)
        mu = np.astype(mu, dtype)
        objects = np.empty(shape, dtype=np.object_)
        for i in np.ndindex(*shape):
            objects[i] = ss.vonmises_fisher(mu[i], kappa[i])
        super().__init__(shape, rvs_shape, dtype, objects)



from __future__ import annotations

from abc import abstractmethod
from typing import Any, TypeVar

from jax.scipy.special import gammaln
from tjax import JaxIntegralArray, JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, integral_ring
from ..parametrization import SimpleDistribution

EP = TypeVar('EP', bound='NBCommonEP[Any]')


@dataclass
class NBCommonNP(NaturalParametrization[EP, JaxRealArray],
                 SimpleDistribution):
    log_not_p: JaxRealArray

    @property
    @override
    def shape(self) -> Shape:
        return self.log_not_p.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return -self._failures() * xp.log1p(-xp.exp(self.log_not_p))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        a = x + self._failures() - 1
        # Return log(a choose x).
        return gammaln(a + 1) - gammaln(x + 1) - gammaln(a - x + 1)

    def _mean(self) -> JaxRealArray:
        xp = self.get_namespace()
        return self._failures() / xp.expm1(-self.log_not_p)

    @abstractmethod
    def _failures(self) -> int | JaxIntegralArray:
        raise NotImplementedError


NP = TypeVar('NP', bound=NBCommonNP[Any])


@dataclass
class NBCommonEP(ExpectationParametrization[NP],
                 SimpleDistribution):
    mean: JaxRealArray

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    # def conjugate_prior_distribution(self, n: JaxRealArray) -> BetaPrimeNP:
    #     return BetaPrimeNP(n * self._failures() * (self.mean, xp.ones_like(self.mean)))

    def _log_not_p(self) -> JaxRealArray:
        xp = self.get_namespace()
        return -xp.log1p(self._failures() / self.mean)

    @abstractmethod
    def _failures(self) -> int | JaxIntegralArray:
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass
from math import prod
from typing import Any

import numpy as np
from numpy.random import Generator
from tjax import NumpyRealArray, NumpyRealNumeric, ShapeLike

from ..tools import join_mappings


@dataclass
class ScipyJointDistribution:
    sub_distributions: dict[str, Any]

    def pdf(self, z: dict[str, Any], out: None = None) -> NumpyRealNumeric:
        joined = join_mappings(sub=self.sub_distributions, z=z)
        return prod(value['sub'].pdf(value['z']) for value in joined.values())

    def rvs(self,
            size: ShapeLike | None = None,
            random_state: Generator | None = None
            ) -> dict[str, Any]:
        return {name: distribution.rvs(size, random_state)
                for name, distribution in self.sub_distributions.items()}

    def entropy(self) -> NumpyRealArray:
        return np.sum([distribution.entropy()
                       for distribution in self.sub_distributions.values()],
                      axis=0)



from __future__ import annotations

from typing import Any, Self

import jax
import numpy as np
from array_api_compat import get_namespace
from jax.scipy import special as jss
from tjax import JaxBooleanArray, JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasConjugatePrior
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, ScalarSupport, boolean_ring, distribution_parameter
from ..parametrization import SimpleDistribution
from .beta import BetaNP


@dataclass
class BernoulliNP(HasEntropyNP['BernoulliEP'],
                  NaturalParametrization['BernoulliEP', JaxRealArray],
                  Samplable,
                  SimpleDistribution):
    """The natural parametrization of the Bernoulli distribution.

    Args:
        log_odds: log(p / (1-p)).
    """
    log_odds: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_odds.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=boolean_ring)

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.logaddexp(self.log_odds, xp.asarray(0.0))

    @override
    def to_exp(self) -> BernoulliEP:
        return BernoulliEP(jss.expit(self.log_odds))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> BernoulliEP:
        return BernoulliEP(x)

    def nat_to_probability(self) -> JaxRealArray:
        xp = self.get_namespace()
        p = jss.expit(self.log_odds)
        final_p = 1.0 - p
        return xp.stack([p, final_p], axis=-1)

    def nat_to_surprisal(self) -> JaxRealArray:
        xp = self.get_namespace()
        total_p = self.nat_to_probability()
        return -xp.log(total_p)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class BernoulliEP(HasEntropyEP[BernoulliNP],
                  HasConjugatePrior,
                  Samplable,
                  ExpectationParametrization[BernoulliNP]):
    """The expectation parametrization of the Bernoulli distribution.

    Args:
        probability: p.
    """
    probability: JaxRealArray = distribution_parameter(ScalarSupport(ring=RealField(
        minimum=0.0, maximum=1.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.probability.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=boolean_ring)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[BernoulliNP]:
        return BernoulliNP

    @override
    def to_nat(self) -> BernoulliNP:
        return BernoulliNP(jss.logit(self.probability))

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxBooleanArray:
        if shape is not None:
            shape += self.shape
        return jax.random.bernoulli(key, self.probability, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> BetaNP:
        xp = self.get_namespace()
        reshaped_n = n[..., np.newaxis]
        return BetaNP(reshaped_n * xp.stack([self.probability, (1.0 - self.probability)], axis=-1))

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, BetaNP)
        a = cp.alpha_minus_one
        xp = get_namespace(a)
        n = xp.sum(a, axis=-1)
        probability = (a[..., 0] / n)
        return (cls(probability), n)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.probability



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Self

from tjax import JaxComplexArray, JaxRealArray

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.multidimensional import Multidimensional
from ..natural_parametrization import NaturalParametrization


class HasConjugatePrior(ExpectationParametrization[Any]):
    @abstractmethod
    def conjugate_prior_distribution(self, n: JaxRealArray
                                     ) -> NaturalParametrization[Any, Any]:
        """The conjugate prior distribution.

        Args:
            n: The nonnegative number of pseudo-observations.  Must have shape == self.shape.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        """Given a conjugate prior distribution, find the distribution and observation count.

        Args:
            cp: The conjugate prior distribution.

        Returns:
            The distribution that gave rise to the conjguate prior, and the observation count.
        """
        raise NotImplementedError

    @abstractmethod
    def conjugate_prior_observation(self) -> JaxComplexArray:
        """An observation of the conjugate prior corresponding to this distribution."""
        raise NotImplementedError


class HasGeneralizedConjugatePrior(HasConjugatePrior, Multidimensional):
    @abstractmethod
    def generalized_conjugate_prior_distribution(self, n: JaxRealArray
                                                 ) -> NaturalParametrization[Any, Any]:
        """A generalization of the conjugate prior distribution.

        Args:
            n: The nonnegative number of pseudo-observations.  Must have
                shape == (*self.shape, self.dimensions()).
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Self

import jax
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasConjugatePrior
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter, negative_support, positive_support
from .gamma import GammaNP


@dataclass
class ExponentialNP(HasEntropyNP['ExponentialEP'],
                    Samplable,
                    NaturalParametrization['ExponentialEP', JaxRealArray]):
    """The natural parametrization of the exponential distribution.

    Args:
        negative_rate: -lambda where lambda is the rate.
    """
    negative_rate: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_rate.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return -xp.log(-self.negative_rate)

    @override
    def to_exp(self) -> ExponentialEP:
        return ExponentialEP(-1.0 / self.negative_rate)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> ExponentialEP:
        return ExponentialEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return -jax.random.exponential(key, shape) / self.negative_rate


@dataclass
class ExponentialEP(HasEntropyEP[ExponentialNP],
                    HasConjugatePrior,
                    Samplable,
                    ExpectationParametrization[ExponentialNP]):
    """The expectation parametrization of the exponential distribution.

    Args:
        mean: E(x) = 1/lambda where lambda is the rate.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ExponentialNP]:
        return ExponentialNP

    @override
    def to_nat(self) -> ExponentialNP:
        return ExponentialNP(-1.0 / self.mean)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.exponential(key, shape) * self.mean

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> GammaNP:
        return GammaNP(-n / self.mean, n)

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, GammaNP)
        n = cp.shape_minus_one
        return (cls(-n / cp.negative_rate), n)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar, final

import jax.numpy as jnp
import pytest
from numpy.random import Generator
from tjax import JaxComplexArray, NumpyComplexArray, Shape
from typing_extensions import override

from efax import ExpectationParametrization, NaturalParametrization, Structure, SubDistributionInfo

NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])
EP = TypeVar('EP', bound=ExpectationParametrization[Any])
Domain = TypeVar('Domain', bound=NumpyComplexArray | dict[str, Any])


class DistributionInfo(Generic[NP, EP, Domain]):
    def __init__(self, dimensions: int = 1, safety: float = 0.0) -> None:
        super().__init__()
        self.dimensions = dimensions
        self.safety = safety

    def exp_to_scipy_distribution(self, p: EP) -> Any:
        """Produce a corresponding scipy distribution from expectation parameters.

        Args:
            p: Expectation parameters.
        """
        return self.nat_to_scipy_distribution(p.to_nat())

    def nat_to_scipy_distribution(self, q: NP) -> Any:
        """Produce a corresponding scipy distribution from natural parameters.

        Args:
            q: Natural parameters.
        """
        return self.exp_to_scipy_distribution(q.to_exp())

    @final
    def exp_parameter_generator(self, rng: Generator, shape: Shape) -> EP:
        """Generate expectation parameters.

        Defaults to converting values generated by nat_parameter_generator.
        """
        return self.nat_parameter_generator(rng, shape).to_exp()

    @final
    def nat_parameter_generator(self, rng: Generator, shape: Shape) -> NP:
        """Generate natural parameters."""
        return self.nat_structure().generate_random(jnp, rng, shape, self.safety)

    def scipy_to_exp_family_observation(self, x: Domain) -> JaxComplexArray | dict[str, Any]:
        """The observation that's expected by the exponential family.

        Args:
            x: The observation that's produced by the scipy distribution.
        """
        return jnp.asarray(x)

    def exp_structure(self) -> Structure[EP]:
        return Structure([SubDistributionInfo((), self.exp_class(), self.dimensions, [])])

    def nat_structure(self) -> Structure[NP]:
        return Structure([SubDistributionInfo((), self.nat_class(), self.dimensions, [])])

    def exp_class(self) -> type[EP]:
        raise NotImplementedError

    def nat_class(self) -> type[NP]:
        raise NotImplementedError

    @classmethod
    def name(cls) -> str:
        return cls.__name__.removesuffix('Info')

    @classmethod
    def tests_selected(cls, distribution_name: str | None) -> bool:
        return distribution_name is None or cls.name() == distribution_name

    @classmethod
    def skip_if_deselected(cls, distribution_name: str | None) -> None:
        if not cls.tests_selected(distribution_name):
            pytest.skip(f"Deselected {cls.name()}")

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        if (cls.exp_to_scipy_distribution is DistributionInfo.exp_to_scipy_distribution
                and cls.nat_to_scipy_distribution is DistributionInfo.nat_to_scipy_distribution):
            raise TypeError

        for method in ('exp_parameter_generator', 'nat_parameter_generator',
                       'scipy_to_exp_family_observation'):
            old_method = getattr(cls, method)

            def new_method(*args: Any,
                           old_method: Callable[..., Any] = old_method,
                           **kwargs: Any) -> Any:
                return old_method(*args, **kwargs)

            setattr(cls, method, new_method)



from __future__ import annotations

from typing import Any

from array_api_compat import get_namespace
from jax.random import gamma as random_gamma
from jax.scipy import special as jss
from tjax import JaxRealArray, KeyArray, Shape, inverse_softplus, softplus
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import (RealField, ScalarSupport, distribution_parameter, negative_support,
                         positive_support)
from ..parametrization import SimpleDistribution


@dataclass
class GammaNP(HasEntropyNP['GammaEP'],
              Samplable,
              NaturalParametrization['GammaEP', JaxRealArray],
              SimpleDistribution):
    """The natural parametrization of the Gamma distribution.

    Args:
        negative_rate: The negative rate.
        shape_minus_one: The shape minus one.
    """
    negative_rate: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))
    shape_minus_one: JaxRealArray = distribution_parameter(ScalarSupport(
        ring=RealField(minimum=-1.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_rate.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        shape = self.shape_minus_one + 1.0
        return jss.gammaln(shape) - shape * xp.log(-self.negative_rate)

    @override
    def to_exp(self) -> GammaEP:
        xp = self.get_namespace()
        shape = self.shape_minus_one + 1.0
        return GammaEP(-shape / self.negative_rate,
                       jss.digamma(shape) - xp.log(-self.negative_rate))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> GammaEP:
        xp = get_namespace(x)
        return GammaEP(x, xp.log(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        return -random_gamma(key, self.shape_minus_one + 1.0, shape) / self.negative_rate

    def to_var(self) -> GammaVP:
        rate = -self.negative_rate
        shape = self.shape_minus_one + 1.0
        mean = shape / rate
        variance = mean / rate
        return GammaVP(mean, variance)


@dataclass
class GammaEP(HasEntropyEP[GammaNP],
              Samplable,
              ExpToNat[GammaNP],
              ExpectationParametrization[GammaNP],
              SimpleDistribution):
    """The expectation parametrization of the Gamma distribution.

    Args:
        mean: The mean: E(x).
        mean_log: The mean of the log: E(log(x)).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))
    mean_log: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[GammaNP]:
        return GammaNP

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)

    @override
    def search_to_natural(self, search_parameters: JaxRealArray) -> GammaNP:
        shape = softplus(search_parameters[..., 0])
        rate = shape / self.mean
        return GammaNP(-rate, shape - 1.0)

    @override
    def search_gradient(self, search_parameters: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace()
        shape = softplus(search_parameters[..., 0])
        log_mean_minus_mean_log = xp.log(self.mean) - self.mean_log
        return (log_mean_minus_mean_log - xp.log(shape) + jss.digamma(shape))[..., xp.newaxis]
        # gradient is 1.0 / shape - jss.polygamma(1, shape)
        # where polygamma(1) is trigamma

    @override
    def initial_search_parameters(self) -> JaxRealArray:
        xp = self.get_namespace()
        log_mean_minus_mean_log = xp.log(self.mean) - self.mean_log
        initial_shape: JaxRealArray = (
            (3.0 - log_mean_minus_mean_log
             + xp.sqrt((log_mean_minus_mean_log - 3.0) ** 2 + 24.0 * log_mean_minus_mean_log))
            / (12.0 * log_mean_minus_mean_log))
        return inverse_softplus(initial_shape)[..., xp.newaxis]


@dataclass
class GammaVP(Samplable, SimpleDistribution):
    """The variance parametrization of the Gamma distribution.

    Args:
        mean: The mean.
        variance: The variance.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))
    variance: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    def to_nat(self) -> GammaNP:
        # mean = s/r
        # var = s/r^2
        rate = self.mean / self.variance
        shape = self.mean * rate
        return GammaNP(-rate, shape - 1.0)

    def to_exp(self) -> GammaEP:
        return self.to_nat().to_exp()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)



from __future__ import annotations

import math
from typing import Any

from array_api_compat import get_namespace
from jax.random import chisquare
from jax.scipy import special as jss
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, ScalarSupport, distribution_parameter
from ..parametrization import SimpleDistribution


@dataclass
class ChiSquareNP(HasEntropyNP['ChiSquareEP'],
                  Samplable,
                  NaturalParametrization['ChiSquareEP', JaxRealArray]):
    """The natural parameters of the chi-square distribution with k degrees of freedom.

    This is the gamma distribution with shape k/2 and rate 1/2.

    Args:
        k_over_two_minus_one: k/2 - 1.
    """
    k_over_two_minus_one: JaxRealArray = distribution_parameter(ScalarSupport(
        ring=RealField(minimum=-1.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.k_over_two_minus_one.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        k_over_two = self.k_over_two_minus_one + 1.0
        return jss.gammaln(k_over_two) - k_over_two * math.log(0.5)

    @override
    def to_exp(self) -> ChiSquareEP:
        k_over_two = self.k_over_two_minus_one + 1.0
        return ChiSquareEP(jss.digamma(k_over_two) - math.log(0.5))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return -x * 0.5

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> ChiSquareEP:
        xp = get_namespace(x)
        return ChiSquareEP(xp.log(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        degrees_of_freedom = (self.k_over_two_minus_one + 1.0) * 2.0
        return chisquare(key, degrees_of_freedom, shape)


# The ExpToNat mixin can be circumvented if the inverse of the digamma function were added to JAX.
@dataclass
class ChiSquareEP(HasEntropyEP[ChiSquareNP],
                  Samplable,
                  ExpToNat[ChiSquareNP],
                  ExpectationParametrization[ChiSquareNP],
                  SimpleDistribution):
    """The expectation parameters of the chi-square distribution with k degrees of freedom.

    This is the gamma distribution with shape k/2 and rate 1/2.

    Args:
        mean_log: E(log(x)).
    """
    mean_log: JaxRealArray = distribution_parameter(ScalarSupport())

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ChiSquareNP]:
        return ChiSquareNP

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_log.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        q = self.to_nat()
        k_over_two = q.k_over_two_minus_one + 1.0
        return -k_over_two

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)



from __future__ import annotations

import numpy as np
import scipy.special
import scipy.stats as ss
from numpy.random import Generator
from tjax import NumpyComplexArray, NumpyRealArray, NumpyRealNumeric, ShapeLike

scipy_beta = scipy.special.beta


class ScipyGeometric:
    """This class repairs Geometric."""
    def __init__(self, p: NumpyRealArray) -> None:
        super().__init__()
        self.geom = ss.geom(p)

    def pmf(self, z: NumpyRealNumeric, out: None = None) -> NumpyRealArray:
        return self.geom.pmf(z)  # pyright: ignore

    def rvs(self,
            size: ShapeLike | None = None,
            random_state: Generator | None = None
            ) -> NumpyComplexArray:
        return np.asarray(self.geom.rvs(size=size, random_state=random_state))

    def entropy(self) -> NumpyRealArray:
        return np.asarray(self.geom.entropy())



from __future__ import annotations

from typing import Any, Self

import jax
from jax.scipy import special as jss
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.conjugate_prior import HasConjugatePrior
from ..interfaces.samplable import Samplable
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter, integral_ring, positive_support
from ..parametrization import SimpleDistribution
from .gamma import GammaNP


@dataclass
class PoissonNP(Samplable,
                NaturalParametrization['PoissonEP', JaxRealArray],
                SimpleDistribution):
    """The natural parametrization of the Poisson distribution.

    Args:
        log_mean: log(E(x)).
    """
    log_mean: JaxRealArray = distribution_parameter(ScalarSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.log_mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.exp(self.log_mean)

    @override
    def to_exp(self) -> PoissonEP:
        xp = self.get_namespace()
        return PoissonEP(xp.exp(self.log_mean))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        return -jss.gammaln(x + 1)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> PoissonEP:
        return PoissonEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class PoissonEP(HasConjugatePrior,
                Samplable,
                ExpectationParametrization[PoissonNP],
                SimpleDistribution):
    """The expectation parametrization of the Poisson distribution.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=integral_ring)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[PoissonNP]:
        return PoissonNP

    @override
    def to_nat(self) -> PoissonNP:
        xp = self.get_namespace()
        return PoissonNP(xp.log(self.mean))

    # The expected_carrier_measure is -exp(-mean) * sum over k from zero to infinity of
    #   lambda ** k * log(k!) / k! = lambda ** k * log Gamma(k+1) / Gamma(k+1)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.poisson(key, self.mean, shape)

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> GammaNP:
        return GammaNP(-n, n * self.mean)

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any]
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, GammaNP)
        n = -cp.negative_rate
        mean = cp.shape_minus_one / n
        return (cls(mean), n)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean



from __future__ import annotations

import math
from typing import Any

import jax
from array_api_compat import get_namespace
from tjax import JaxArray, JaxComplexArray, JaxRealArray, KeyArray, Shape, abs_square
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import (ComplexField, RealField, ScalarSupport, Support, complex_field,
                          distribution_parameter)
from ...parametrization import SimpleDistribution


@dataclass
class ComplexNormalNP(HasEntropyNP['ComplexNormalEP'],
                      Samplable,
                      NaturalParametrization['ComplexNormalEP', JaxComplexArray],
                      SimpleDistribution):
    """The natural parametrization of the complex normal distribution.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_precision: -1 / Var(x).
        pseudo_precision: 1 / PVar(x).

    Where the pseudo-variance is
        PVar(x) = E(x^2 - E(x)^2)

    We know that |PVar(x)| < Var(x) where |x| is the modulus.
    """
    mean_times_precision: JaxComplexArray = distribution_parameter(
        ScalarSupport(ring=complex_field))
    negative_precision: JaxRealArray = distribution_parameter(
            ScalarSupport(ring=RealField(maximum=0.0)))
    pseudo_precision: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        _, s, mu = self._r_s_mu()
        det_s = s
        det_h = -self.negative_precision
        return (-abs_square(mu) * self.negative_precision
                - xp.real(xp.square(mu) * self.pseudo_precision)
                + 0.5 * xp.log(det_s)
                - 0.5 * xp.log(det_h)
                + math.log(math.pi))

    @override
    def to_exp(self) -> ComplexNormalEP:
        xp = self.get_namespace()
        r, s, mu = self._r_s_mu()
        u = xp.conj(r * s)
        return ComplexNormalEP(mu, s + abs_square(mu), u + xp.square(mu))

    @override
    def carrier_measure(self, x: JaxComplexArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxComplexArray, **fixed_parameters: Any
                              ) -> ComplexNormalEP:
        xp = get_namespace(x)
        return ComplexNormalEP(x, abs_square(x), xp.square(x))

    def _r_s_mu(self) -> tuple[JaxComplexArray, JaxRealArray, JaxComplexArray]:
        xp = self.get_namespace()
        r = -self.pseudo_precision / self.negative_precision
        s = 1.0 / ((abs_square(r) - 1.0) * self.negative_precision)
        k = self.pseudo_precision / self.negative_precision
        l_eta = 0.5 * self.mean_times_precision / ((abs_square(k) - 1.0) * self.negative_precision)
        mu = xp.conj(l_eta) - xp.conj(self.pseudo_precision / self.negative_precision) * l_eta
        return r, s, mu

    @override
    @classmethod
    def adjust_support(cls, name: str, **kwargs: JaxArray) -> Support:
        if name != 'pseudo_precision':
            return super().adjust_support(name, **kwargs)
        precision = -kwargs['negative_precision']
        xp = get_namespace(precision)
        return ScalarSupport(ring=ComplexField(maximum_modulus=xp.abs(precision)))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class ComplexNormalEP(HasEntropyEP[ComplexNormalNP],
                      Samplable,
                      ExpectationParametrization[ComplexNormalNP],
                      SimpleDistribution):
    """The expectation parametrization of the complex normal distribution.

    Args:
        mean: E(x).
        second_moment: E(x * conjugate(x)).
        pseudo_second_moment: E(x^2).
    """
    mean: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))
    second_moment: JaxRealArray = distribution_parameter(ScalarSupport(ring=RealField(minimum=0.0)))
    pseudo_second_moment: JaxComplexArray = distribution_parameter(
        ScalarSupport(ring=complex_field))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ComplexNormalNP]:
        return ComplexNormalNP

    @override
    def to_nat(self) -> ComplexNormalNP:
        xp = self.get_namespace()
        variance = self.second_moment - abs_square(self.mean)
        pseudo_variance = self.pseudo_second_moment - xp.square(self.mean)

        r = xp.conj(pseudo_variance) / variance
        p_c = variance - xp.conj(r * pseudo_variance)
        p_inv_c = 1.0 / p_c
        precision = xp.real(-p_inv_c)
        pseudo_precision = r * p_inv_c
        mean_times_precision = -2.0 * (precision * xp.conj(self.mean)
                                       + pseudo_precision * self.mean)
        return ComplexNormalNP(mean_times_precision, precision, pseudo_precision)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    @classmethod
    def adjust_support(cls, name: str, **kwargs: JaxArray) -> Support:
        if name != 'pseudo_precision':
            return super().adjust_support(name, **kwargs)
        precision = -kwargs['negative_precision']
        xp = get_namespace(precision)
        return ScalarSupport(ring=ComplexField(maximum_modulus=xp.abs(precision)))

    def _multivariate_normal_cov(self) -> JaxRealArray:
        xp = self.get_namespace()
        variance = self.second_moment - abs_square(self.mean)
        pseudo_variance = self.pseudo_second_moment - xp.square(self.mean)
        xx = variance + xp.real(pseudo_variance)
        xy = xp.imag(pseudo_variance)
        yy = variance - xp.real(pseudo_variance)
        xx_xy = xp.stack([xx, xy], axis=-1)
        yx_yy = xp.stack([xy, yy], axis=-1)
        return 0.5 * xp.stack([xx_xy, yx_yy], axis=-2)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        xp = self.get_namespace()
        mn_mean = xp.stack([xp.real(self.mean), xp.imag(self.mean)], axis=-1)
        mn_cov = self._multivariate_normal_cov()
        mn_sample = jax.random.multivariate_normal(key, mn_mean, mn_cov, shape)
        return mn_sample[..., 0] + 1j * mn_sample[..., 1]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Self, override

from array_api_compat import get_namespace
from tjax import JaxAbstractClass, JaxArray, Shape
from tjax.dataclasses import dataclass

from .parameter import Support


@dataclass
class Distribution(JaxAbstractClass):
    """The Distribution is the base class of all distributions."""
    def __getitem__(self, key: Any) -> Self:
        from .iteration import parameters  # noqa: PLC0415
        from .structure import Structure  # noqa: PLC0415
        parameters_ = {path: value[key] for path, value in parameters(self).items()}
        return Structure.create(self).assemble(parameters_)

    @abstractmethod
    def sub_distributions(self) -> Mapping[str, Distribution]:
        raise NotImplementedError

    @property
    @abstractmethod
    def shape(self) -> Shape:
        raise NotImplementedError

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @classmethod
    def adjust_support(cls, name: str, **kwargs: JaxArray) -> Support:
        from .iteration import support  # noqa: PLC0415
        return support(cls)[name]

    def get_namespace(self, *x: Any) -> Any:
        from .iteration import parameters  # noqa: PLC0415
        values = parameters(self, support=False).values()
        return get_namespace(*values)


@dataclass
class SimpleDistribution(Distribution):
    """A SimpleDistribution has no sub-distributions.

    As a consequence, its domain is a simple support (rather than a dict).
    """
    @override
    def sub_distributions(self) -> Mapping[str, Distribution]:
        return {}

    @classmethod
    @abstractmethod
    def domain_support(cls) -> Support:
        raise NotImplementedError



from __future__ import annotations

from typing import Any

import numpy as np
import optype.numpy as onp
import scipy.special
import scipy.stats as ss
from numpy.random import Generator
from tjax import NumpyComplexArray, NumpyRealArray, ShapeLike
from typing_extensions import override

from .shaped_distribution import ShapedDistribution

scipy_beta = scipy.special.beta


mvd: type = ss._multivariate.dirichlet_frozen  # noqa: SLF001


class ScipyDirichletFixRVsAndPDF(mvd):
    """This class repairs dirichlet.

    See https://github.com/scipy/scipy/issues/6005 and https://github.com/scipy/scipy/issues/6006.
    """
    @override
    def rvs(self,
            size: Any = None,
            random_state: Any = None
            ) -> onp.ArrayND[np.float64]:
        if size is None:
            size = ()
        # This somehow fixes the behaviour of rvs.
        return super().rvs(size=size, random_state=random_state)

    @override
    def pdf(self, x: onp.ToFloatND) -> np.float64:
        x = np.asarray(x)
        if x.ndim == 2:  # noqa: PLR2004
            return np.float64(super().pdf(x.T))
        return np.float64(super().pdf(x))


class ScipyDirichlet(ShapedDistribution[ScipyDirichletFixRVsAndPDF]):
    """This class allows distributions having a non-empty shape."""
    @override
    def __init__(self, alpha: NumpyRealArray) -> None:
        shape = alpha[..., -1].shape
        rvs_shape = (alpha.shape[-1],)
        dtype = alpha.dtype
        objects = np.empty(shape, dtype=ScipyDirichletFixRVsAndPDF)
        for i in np.ndindex(shape):
            objects[i] = ScipyDirichletFixRVsAndPDF(alpha[i])
        super().__init__(shape, rvs_shape, dtype, objects)

    @override
    def pdf(self, x: NumpyComplexArray) -> NumpyRealArray:
        x = np.astype(x, np.float64)
        y = np.sum(x, axis=-1)
        if not np.all(np.isclose(y, np.ones(y.shape), atol=1e-5, rtol=0)):
            raise ValueError
        return super().pdf(x)


class ScipyGeneralizedDirichlet:
    @override
    def __init__(self, alpha: NumpyRealArray, beta: NumpyRealArray) -> None:
        super().__init__()
        self.alpha = alpha
        self.beta = beta

    def pdf(self, x: NumpyRealArray) -> NumpyRealArray:
        alpha_roll = np.roll(self.alpha, -1, axis=-1)
        alpha_roll[..., -1] = 0.0
        gamma = -np.diff(self.beta, append=1.0) - alpha_roll
        cs_x = np.cumsum(x, axis=-1)
        terms = x ** (self.alpha - 1.0) * (1.0 - cs_x) ** gamma / scipy_beta(self.alpha, self.beta)
        return np.prod(terms, axis=-1)

    def rvs(self, size: ShapeLike = (), random_state: Generator | None = None) -> NumpyComplexArray:
        sample_size: tuple[int, ...] = (((size,) if isinstance(size, int) else tuple(size))
                                        + self.alpha.shape)
        if random_state is None:
            random_state = np.random.default_rng()
        dimensions = self.alpha.shape[-1]
        beta_samples = random_state.beta(self.alpha, self.beta, size=sample_size)
        q = np.zeros(beta_samples.shape[:-1])
        for i in range(dimensions):
            beta_samples[..., i] *= 1 - q
            q += beta_samples[..., i]
        return beta_samples

    def entropy(self) -> NumpyRealArray:
        raise RuntimeError



from __future__ import annotations

from abc import abstractmethod

from tjax import JaxArray, KeyArray, Shape

from ..parametrization import SimpleDistribution


class Samplable(SimpleDistribution):
    @abstractmethod
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxArray:
        raise NotImplementedError



from __future__ import annotations

from typing import Any

from array_api_compat import get_namespace
from tjax import JaxRealArray
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import SimplexSupport
from .dirichlet_common import DirichletCommonEP, DirichletCommonNP


@dataclass
class DirichletNP(DirichletCommonNP['DirichletEP'],
                  NaturalParametrization['DirichletEP', JaxRealArray]):
    """The natural parametrization of the Dirichlet distribution.

    Args:
        alpha_minus_one: The shape parameters alpha.
    """
    @override
    @classmethod
    def domain_support(cls) -> SimplexSupport:
        return SimplexSupport()

    @override
    def to_exp(self) -> DirichletEP:
        return DirichletEP(self._exp_helper())

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> DirichletEP:
        xp = get_namespace(x)
        one_minus_total_x = 1.0 - xp.sum(x, axis=-1, keepdims=True)
        return DirichletEP(xp.concat((xp.log(x), xp.log(one_minus_total_x)), axis=-1))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:len(x.shape) - 1])


@dataclass
class DirichletEP(DirichletCommonEP[DirichletNP],
                  ExpectationParametrization[DirichletNP]):
    """The expectation parametrization of the Dirichlet distribution.

    Args:
        mean_log_probability: An array with final dimension [E(log(x_i))]_i.
    """
    @override
    @classmethod
    def domain_support(cls) -> SimplexSupport:
        return SimplexSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[DirichletNP]:
        return DirichletNP



"""These tests are related to entropy."""
from __future__ import annotations

from functools import partial
from typing import Any

import jax.numpy as jnp
import pytest
from jax import grad, tree, value_and_grad
from jax.test_util import check_grads
from numpy.random import Generator
from rich.console import Console
from tjax import JaxRealArray, assert_tree_allclose, print_generic

from efax import Flattener, GammaEP, GammaVP, HasEntropy

from .create_info import BetaInfo, DirichletInfo
from .distribution_info import DistributionInfo


def sum_entropy(flattened: JaxRealArray, flattener: Flattener[Any]) -> JaxRealArray:
    x = flattener.unflatten(flattened)
    return jnp.sum(x.entropy())


def all_finite(some_tree: Any, /) -> bool:
    return tree.all(tree.map(lambda x: jnp.all(jnp.isfinite(x)), some_tree))


def check_entropy_gradient(distribution: HasEntropy, /) -> None:
    flattener, flattened = Flattener.flatten(distribution, map_to_plane=False)
    calculated_gradient = grad(sum_entropy)(flattened, flattener)
    if not all_finite(calculated_gradient):
        indices = jnp.argwhere(jnp.isnan(calculated_gradient))
        bad_distributions = [distribution[tuple(index)[:-1]] for index in indices]
        console = Console()
        with console.capture() as capture:
            print_generic(bad_distributions, console=console)
        msg = f"Non-finite gradient found for distributions: {capture.get()}"
        raise AssertionError(msg)
    check_grads(partial(sum_entropy, flattener=flattener), (flattened,), order=1, atol=1e-4,
                rtol=1e-2)


def test_nat_entropy_gradient(generator: Generator,
                              entropy_distribution_info: DistributionInfo[Any, Any, Any],
                              ) -> None:
    if isinstance(entropy_distribution_info, BetaInfo | DirichletInfo):
        pytest.skip()
    shape = (7, 13)
    nat_parameters = entropy_distribution_info.nat_parameter_generator(generator, shape=shape)
    check_entropy_gradient(nat_parameters)


def test_exp_entropy_gradient(generator: Generator,
                              entropy_distribution_info: DistributionInfo[Any, Any, Any],
                              ) -> None:
    if isinstance(entropy_distribution_info, BetaInfo | DirichletInfo):
        pytest.skip()
    shape = (7, 13)
    exp_parameters = entropy_distribution_info.exp_parameter_generator(generator, shape=shape)
    check_entropy_gradient(exp_parameters)


def test_gamma_vp_entropy_gradient() -> None:
    def f(flattened: JaxRealArray, flattener: Flattener[GammaEP]) -> JaxRealArray:
        x_e = flattener.unflatten(flattened)
        return x_e.entropy()

    def g(flattened: JaxRealArray, flattener: Flattener[GammaEP]) -> JaxRealArray:
        x_e = flattener.unflatten(flattened)
        x_n = x_e.to_nat()
        return x_n.entropy()

    o = jnp.ones(())
    x = GammaVP(o, o * 0.25)
    x_e = x.to_exp()
    flattener, flattened = Flattener.flatten(x_e, map_to_plane=False)
    f_value, f_gradient = value_and_grad(f)(flattened, flattener)
    g_value, g_gradient = value_and_grad(g)(flattened, flattener)
    assert_tree_allclose(f_value, g_value)
    assert_tree_allclose(f_gradient, g_gradient)



from __future__ import annotations

import math
from typing import Any, cast

import jax
from tjax import JaxRealArray, KeyArray, Shape, matrix_dot_product, matrix_vector_mul, outer_product
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import SymmetricMatrixSupport, VectorSupport, distribution_parameter


@dataclass
class MultivariateNormalNP(HasEntropyNP['MultivariateNormalEP'],
                           NaturalParametrization['MultivariateNormalEP', JaxRealArray],
                           Multidimensional,
                           Samplable):
    """The natural parametrization of the multivariate normal distribution.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_half_precision: -0.5 / Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(SymmetricMatrixSupport(
        negative_semidefinite=True))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        eta = self.mean_times_precision
        h_inv = xp.linalg.inv(self.negative_half_precision)
        a = matrix_dot_product(h_inv, outer_product(eta, eta))
        _, ld = xp.linalg.slogdet(-self.negative_half_precision)
        return -0.25 * a - 0.5 * ld + 0.5 * self.dimensions() * math.log(math.pi)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_variance_parametrization().sample(key, shape)

    def variance(self) -> JaxRealArray:
        xp = self.get_namespace()
        h_inv: JaxRealArray = xp.linalg.inv(self.negative_half_precision)
        return -0.5 * h_inv

    def mean(self) -> JaxRealArray:
        variance = self.variance()
        return matrix_vector_mul(variance, self.mean_times_precision)

    def to_variance_parametrization(self) -> MultivariateNormalVP:
        variance = self.variance()
        mean = matrix_vector_mul(variance, self.mean_times_precision)
        return MultivariateNormalVP(mean, variance)

    @override
    def to_exp(self) -> MultivariateNormalEP:
        xp = self.get_namespace()
        h_inv = xp.linalg.inv(self.negative_half_precision)
        h_inv = cast('JaxRealArray', h_inv)
        h_inv_times_eta = matrix_vector_mul(h_inv, self.mean_times_precision)
        mean = -0.5 * h_inv_times_eta
        second_moment = 0.25 * outer_product(h_inv_times_eta, h_inv_times_eta) - 0.5 * h_inv
        return MultivariateNormalEP(mean, second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateNormalEP:
        return MultivariateNormalEP(x, outer_product(x, x))

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateNormalEP(HasEntropyEP[MultivariateNormalNP],
                           ExpectationParametrization[MultivariateNormalNP], Multidimensional,
                           Samplable):
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    second_moment: JaxRealArray = distribution_parameter(SymmetricMatrixSupport(
        positive_semidefinite=True))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateNormalNP]:
        return MultivariateNormalNP

    @override
    def to_nat(self) -> MultivariateNormalNP:
        xp = self.get_namespace()
        precision = xp.linalg.inv(self.variance())
        precision = cast('JaxRealArray', precision)
        mean_times_precision = matrix_vector_mul(precision, self.mean)
        return MultivariateNormalNP(mean_times_precision, -0.5 * precision)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_variance_parametrization().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> JaxRealArray:
        return self.second_moment - outer_product(self.mean, self.mean)

    def to_variance_parametrization(self) -> MultivariateNormalVP:
        return MultivariateNormalVP(self.mean, self.variance())


@dataclass
class MultivariateNormalVP(Samplable, Multidimensional):
    """The expectation parametrization of the multivariate normal distribution.

    Args:
        mean: E(x).
        variance: Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(SymmetricMatrixSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        return jax.random.multivariate_normal(key, self.mean, self.variance, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def to_exp(self) -> MultivariateNormalEP:
        second_moment = self.variance + outer_product(self.mean, self.mean)
        return MultivariateNormalEP(self.mean, second_moment)



from __future__ import annotations

import math
from typing import Any

import numpy as np
from array_api_compat import get_namespace
from jax.random import rayleigh
from tjax import Array, JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..mixins.transformed_parametrization import (TransformedExpectationParametrization,
                                                  TransformedNaturalParametrization)
from ..parameter import ScalarSupport, distribution_parameter, negative_support, positive_support
from .exponential import ExponentialEP, ExponentialNP


@dataclass
class RayleighNP(Samplable,
                 HasEntropyNP['RayleighEP'],
                 TransformedNaturalParametrization[ExponentialNP, ExponentialEP, 'RayleighEP',
                                                   JaxRealArray]):
    """The natural parametrization of the Rayleigh distribution.

    Args:
        eta: -1 / (2 * sigma^2).
    """
    eta: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    @classmethod
    def expectation_parametrization_cls(cls) -> type[RayleighEP]:
        return RayleighEP

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ExponentialNP]:
        return ExponentialNP

    @override
    def base_distribution(self) -> ExponentialNP:
        return ExponentialNP(self.eta)

    @override
    @classmethod
    def create_expectation(cls, expectation_parametrization: ExponentialEP) -> RayleighEP:
        return RayleighEP(expectation_parametrization.mean)

    @override
    @classmethod
    def sample_to_base_sample(cls, x: Array, **fixed_parameters: Any
                              ) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.square(x)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.log(x) + math.log(2.0)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        xp = self.get_namespace()
        if shape is not None:
            shape += self.shape
        sigma = xp.sqrt(-0.5 / self.eta)
        return rayleigh(key, sigma, shape)


@dataclass
class RayleighEP(Samplable,
                 HasEntropyEP[RayleighNP],
                 TransformedExpectationParametrization[ExponentialEP, ExponentialNP, RayleighNP]):
    """The expectation parametrization of the Rayleigh distribution.

    Args:
        chi: 2 * sigma^2.
    """
    chi: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[RayleighNP]:
        return RayleighNP

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ExponentialEP]:
        return ExponentialEP

    @override
    def base_distribution(self) -> ExponentialEP:
        return ExponentialEP(self.chi)

    @override
    def create_natural(self, natural_parametrization: ExponentialNP) -> RayleighNP:
        return RayleighNP(natural_parametrization.negative_rate)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return 0.5 * xp.log(self.chi * 0.5) + (1.5 * math.log(2.0) - 0.5 * np.euler_gamma)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        xp = self.get_namespace()
        if shape is not None:
            shape += self.shape
        sigma = xp.sqrt(0.5 * self.chi)
        return rayleigh(key, sigma, shape)



"""Maximum likelihood estimation.

This example is based on section 1.3.2 from expfam.pdf, entitled Maximum likelihood estimation.

Suppose you have some samples from a distribution family with unknown parameters, and you want to
estimate the maximum likelihood parmaters of the distribution.
"""
import jax.numpy as jnp
from jax.random import key

from efax import DirichletEP, DirichletNP, MaximumLikelihoodEstimator, parameter_mean

# Consider a Dirichlet distribution with a given alpha.
alpha = jnp.asarray([2.0, 3.0, 4.0])
source_distribution = DirichletNP(alpha_minus_one=alpha - 1.0)

# Let's sample from it.
n_samples = 10000
key_a = key(123)
samples = source_distribution.sample(key_a, (n_samples,))

# Now, let's find the maximum likelihood Dirichlet distribution that fits it.
# First, convert the samples to their sufficient statistics.
estimator = MaximumLikelihoodEstimator.create_simple_estimator(DirichletEP)
ss = estimator.sufficient_statistics(samples)
# ss has type DirichletEP.  This is similar to the conjguate prior of the Dirichlet distribution.

# Take the mean over the first axis.
ss_mean = parameter_mean(ss, axis=0)  # ss_mean also has type DirichletEP.

# Convert this back to the natural parametrization.
estimated_distribution = ss_mean.to_nat()
print(estimated_distribution.alpha_minus_one + 1.0)  # [1.9849904 3.0065458 3.963935 ]  # noqa: T201



"""Cross-entropy.

This example is based on section 1.4.1 from expfam.pdf, entitled Information theoretic statistics.
"""
import jax.numpy as jnp
from tjax import print_generic

from efax import BernoulliEP, BernoulliNP

# p is the expectation parameters of three Bernoulli distributions having probabilities 0.4, 0.5,
# and 0.6.
p = BernoulliEP(jnp.asarray([0.4, 0.5, 0.6]))

# q is the natural parameters of three Bernoulli distributions having log-odds 0, which is
# probability 0.5.
q = BernoulliNP(jnp.zeros(3))

print_generic(p.cross_entropy(q))
# Jax Array (3,) float32
# └──  0.6931 │ 0.6931 │ 0.6931

# q2 is natural parameters of Bernoulli distributions having a probability of 0.3.
p2 = BernoulliEP(0.3 * jnp.ones(3))
q2 = p2.to_nat()

print_generic(p.cross_entropy(q2))
# Jax Array (3,) float32
# └──  0.6956 │ 0.7803 │ 0.8651
# A Bernoulli distribution with probability 0.3 predicts a Bernoulli observation with probability
# 0.4 better than the other observations.



from __future__ import annotations

import math
from typing import Any

from array_api_compat import get_namespace
from jax.scipy import special as jss
from tjax import Array, JaxRealArray
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..mixins.transformed_parametrization import (TransformedExpectationParametrization,
                                                  TransformedNaturalParametrization)
from ..parameter import RealField, ScalarSupport, distribution_parameter
from ..parametrization import SimpleDistribution
from .chi_square import ChiSquareEP, ChiSquareNP


@dataclass
class ChiNP(HasEntropyNP['ChiEP'],
            TransformedNaturalParametrization[ChiSquareNP, ChiSquareEP, 'ChiEP', JaxRealArray],
            SimpleDistribution):
    """The natural parametrization of the chi distribution.

    Args:
        k_over_two_minus_one: k/2 - 1 where k is the shape parameter.
    """
    k_over_two_minus_one: JaxRealArray = distribution_parameter(ScalarSupport(
        ring=RealField(minimum=-1.0)))

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=RealField(minimum=0.0))

    @override
    @classmethod
    def expectation_parametrization_cls(cls) -> type[ChiEP]:
        return ChiEP

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ChiSquareNP]:
        return ChiSquareNP

    @override
    def base_distribution(self) -> ChiSquareNP:
        return ChiSquareNP(self.k_over_two_minus_one)

    @override
    @classmethod
    def create_expectation(cls, expectation_parametrization: ChiSquareEP) -> ChiEP:
        return ChiEP(expectation_parametrization.mean_log)

    @override
    @classmethod
    def sample_to_base_sample(cls, x: Array, **fixed_parameters: Any) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.square(x)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.log(2.0 * x) - xp.square(x) * 0.5


@dataclass
class ChiEP(HasEntropyEP[ChiNP],
            TransformedExpectationParametrization[ChiSquareEP, ChiSquareNP, ChiNP],
            SimpleDistribution):
    """The expectation parametrization of the chi distribution.

    Args:
        mean_log: E(log(x)).
    """
    mean_log: JaxRealArray = distribution_parameter(ScalarSupport())

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=RealField(minimum=0.0))

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ChiNP]:
        return ChiNP

    @override
    @classmethod
    def base_distribution_cls(cls) -> type[ChiSquareEP]:
        return ChiSquareEP

    @override
    def base_distribution(self) -> ChiSquareEP:
        return ChiSquareEP(self.mean_log)

    @override
    def create_natural(self, natural_parametrization: ChiSquareNP) -> ChiNP:
        return ChiNP(natural_parametrization.k_over_two_minus_one)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        q = self.to_nat()
        k_over_two = q.k_over_two_minus_one + 1.0
        return -1.0 * k_over_two + 0.5 * jss.digamma(k_over_two) + 1.5 * math.log(2.0)



from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from functools import reduce
from itertools import starmap
from typing import TYPE_CHECKING, Any, TypeVar

from array_api_compat import get_namespace
from tensorflow_probability.substrates import jax as tfp
from tjax import JaxComplexArray, JaxRealArray

from .iteration import parameters
from .parametrization import Distribution
from .structure import Structure
from .types import Axis


def parameter_dot_product(x: NaturalParametrization[Any, Any], y: Any, /) -> JaxRealArray:
    """Return the vectorized dot product over all of the variable parameters."""
    def dotted_fields() -> Iterable[JaxRealArray]:
        xs = parameters(x, fixed=False, support=True).values()
        ys = parameters(y, fixed=False, support=True).values()

        for (x_value, x_support), (y_value, y_support) in zip(xs, ys, strict=True):
            axes = x_support.axes()
            assert y_support.axes() == axes
            yield _parameter_dot_product(x_value, y_value, axes)

    dotted_fields_list = list(dotted_fields())
    xp = get_namespace(*dotted_fields_list)
    return reduce(xp.add, dotted_fields_list)


T = TypeVar('T', bound=Distribution)


def parameter_mean(x: T, /, *, axis: Axis | None = None) -> T:
    """Return the mean of the parameters (fixed and variable)."""
    xp = x.get_namespace()
    structure = Structure.create(x)
    p = parameters(x, support=False)
    q = {path: xp.mean(value, axis=axis) for path, value in p.items()}
    return structure.assemble(q)


def parameter_map(operation: Callable[..., JaxComplexArray],
                  x: T,
                  /,
                  *ys: Distribution
                  ) -> T:
    """Return a new distribution created by operating on the variable fields of the inputs."""
    paths = parameters(x, fixed=False).keys()
    iterators = [parameters(y, fixed=False).values()  # type: ignore[call-overload]
                 for y in (x, *ys)]
    final_values = list(starmap(operation, zip(*iterators, strict=True)))
    operated_fields = dict(zip(paths, final_values, strict=True))

    fixed_parameters = parameters(x, fixed=True)
    return Structure.create(x).assemble({**fixed_parameters, **operated_fields})


_T = TypeVar('_T')
_V = TypeVar('_V')


def join_mappings(**field_to_map: Mapping[_T, _V]) -> dict[_T, dict[str, _V]]:
    """Joins multiple mappings together using their common keys.

    >>> user_scores = {'elliot': 50, 'claris': 60}
    >>> user_times = {'elliot': 30, 'claris': 40}
    >>> join_mappings(score=user_scores, time=user_times)
    {'elliot': {'score': 50, 'time': 30}, 'claris': {'score': 60, 'time': 40}}
    """
    retval = defaultdict[_T, dict[str, _V]](dict)
    for field_name, mapping in field_to_map.items():
        for key, value in mapping.items():
            retval[key][field_name] = value
    return dict(retval)


iv_ratio = tfp.math.bessel_iv_ratio
log_ive = tfp.math.log_bessel_ive


# Private functions --------------------------------------------------------------------------------
def _parameter_dot_product(x: JaxComplexArray, y: JaxComplexArray, n_axes: int) -> JaxRealArray:
    """Returns the real component of the dot product of the final n_axes axes of two arrays."""
    xp = get_namespace(x, y)
    axes = tuple(range(-n_axes, 0))
    return xp.sum(xp.real(x * y), axis=axes)


if TYPE_CHECKING:
    from .natural_parametrization import NaturalParametrization
